"""Class implementation for the scale-y animation value.
"""

from typing import Generic
from typing import TypeVar
from typing import Union

from apysc._animation.animation_base import AnimationBase
from apysc._animation.easing import Easing
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)


class AnimationScaleYFromCenter(AnimationBase[_T], Generic[_T]):
    """
    The animation class for a scale-y.
    """

    _scale_y_from_center: Number
    _before_scale_y_from_center: Number
    _scale_y_from_center_diff_ratio: Number

    def __init__(
            self,
            *,
            target: _T,
            scale_y_from_center: Union[float, Number],
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> None:
        """
        The animation class for a scale-y.

        Parameters
        ----------
        target : SkewYInterface
            A target instance of the animation target
            (e.g., `Rectangle` instance).
        scale_y_from_center : float or Number
            The final scale-y of the animation.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Raises
        ------
        TypeError
            If a specified target is not a ScaleYFromCenterInterface
            instance.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=AnimationScaleYFromCenter):
            from apysc._converter import to_apysc_val_from_builtin
            from apysc._display.scale_y_from_center_interface import \
                ScaleYFromCenterInterface
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            variable_name: str = expression_variables_util.\
                get_next_variable_name(
                    type_name=var_names.ANIMATION_SCALE_Y_FROM_CENTER)
            target_: VariableNameInterface = target
            if isinstance(target_, ScaleYFromCenterInterface):
                target_._initialize_scale_y_from_center_if_not_initialized()
                self._before_scale_y_from_center = \
                    target_._scale_y_from_center
            else:
                raise TypeError(
                    'Specified `target` argument is not a '
                    f'ScaleYFromCenterInterface instance: {type(target_)}')
            self._scale_y_from_center = to_apysc_val_from_builtin.\
                get_copied_number_from_builtin_val(
                    float_or_num=scale_y_from_center)
            one: ap.Number = ap.Number(1.0)
            self._scale_y_from_center_diff_ratio = (
                one / self._before_scale_y_from_center
                * self._scale_y_from_center)
            self._set_basic_animation_settings(
                target=target,
                duration=duration,
                delay=delay,
                easing=easing)
            super(AnimationScaleYFromCenter, self).__init__(
                variable_name=variable_name)

    def _get_animation_func_expression(self) -> str:
        """
        Get a animation function expression.

        Returns
        -------
        expression : str
            Animation function expression.
        """
        from apysc._type import value_util
        diff_ratio_str: str = value_util.get_value_str_for_expression(
            value=self._scale_y_from_center_diff_ratio)
        return f'\n  .scale(1, {diff_ratio_str});'

    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to be inserted into the complete event
        handler's head.

        Returns
        -------
        expression : str
            An expression to be inserted into the complete event
            handler's head.
        """
        from apysc._display.scale_y_from_center_interface import \
            ScaleYFromCenterInterface
        expression: str = ''
        if isinstance(self._target, ScaleYFromCenterInterface):
            self._target._initialize_scale_y_from_center_if_not_initialized()
            expression = (
                f'{self._target._scale_y_from_center.variable_name} = '
                f'{self._scale_y_from_center.variable_name};'
            )
        return expression



"""Class implementation for the animation_line_alpha interface.
"""

from typing import Union

from apysc._animation.animation_interface_base import AnimationInterfaceBase
from apysc._animation.animation_line_alpha import AnimationLineAlpha
from apysc._animation.easing import Easing
from apysc._type.int import Int
from apysc._type.number import Number


class AnimationLineAlphaInterface(AnimationInterfaceBase):

    def animation_line_alpha(
            self,
            alpha: Union[float, Number],
            *,
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> AnimationLineAlpha:
        """
        Set the line alpha animation setting.

        Notes
        -----
        To start this animation, you need to call the `start` method of
        the returned instance.

        Parameters
        ----------
        alpha : float or Number
            The final line alpha of the animation.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Returns
        -------
        animation_line_alpha : AnimationLineAlpha
            Created animation setting instance.

        References
        ----------
        - animation_line_alpha interface document
            - https://simon-ritchie.github.io/apysc/animation_line_alpha.html
        - Animation interfaces duration setting document
            - https://simon-ritchie.github.io/apysc/animation_duration.html
        - Animation interfaces delay setting document
            - https://simon-ritchie.github.io/apysc/animation_delay.html
        - Each animation interface return value document
            - https://bit.ly/2XOoa8w
        - Sequential animation setting document
            - https://simon-ritchie.github.io/apysc/sequential_animation.html
        - animation_parallel interface document
            - https://simon-ritchie.github.io/apysc/animation_parallel.html
        - Easing enum document
            - https://simon-ritchie.github.io/apysc/easing_enum.html
        """
        animation_line_alpha: AnimationLineAlpha = AnimationLineAlpha(
            target=self,
            alpha=alpha,
            duration=duration,
            delay=delay,
            easing=easing)
        return animation_line_alpha



"""Class implementation for the rotation around the given
point animation value.
"""

from typing import Generic
from typing import TypeVar
from typing import Union

from apysc._animation.animation_base import AnimationBase
from apysc._animation.easing import Easing
from apysc._type.int import Int
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)


class AnimationRotationAroundPoint(AnimationBase[_T], Generic[_T]):
    """
    The animation class for a rotation around the given point.
    """

    _rotation_around_point: Int
    _x: Int
    _y: Int
    _before_rotation_around_point: Int
    _rotation_around_point_diff: Int

    def __init__(
            self,
            *,
            target: _T,
            rotation_around_point: Union[int, Int],
            x: Union[int, Int],
            y: Union[int, Int],
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> None:
        """
        The animation class for a rotation around the given point.

        Parameters
        ----------
        target : RotationAroundPointInterface
            A target instance of the animation target
            (e.g., `Rectangle` instance).
        rotation_around_point : int or Int
            The final rotation around the given point of the animation.
        x : int or Int
            X-coordinate.
        y : int or Int
            Y-coordinate.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Raises
        ------
        TypeError
            If a specified target is not a RotationAroundPointInterface
            instance.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=AnimationRotationAroundPoint):
            from apysc._converter import to_apysc_val_from_builtin
            from apysc._display.rotation_around_point_interface import \
                RotationAroundPointInterface
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            variable_name: str = expression_variables_util.\
                get_next_variable_name(
                    type_name=var_names.ANIMATION_ROTATION_AROUND_POINT)
            self._x = to_apysc_val_from_builtin.\
                get_copied_int_from_builtin_val(integer=x)
            self._y = to_apysc_val_from_builtin.\
                get_copied_int_from_builtin_val(integer=y)
            target_: VariableNameInterface = target
            if isinstance(target_, RotationAroundPointInterface):
                target_._initialize_rotation_around_point_if_not_initialized()
                self._before_rotation_around_point = target_.\
                    get_rotation_around_point(x=self._x, y=self._y)
            else:
                raise TypeError(
                    'Specified `target` argument is not a '
                    f'RotationAroundPointInterface instance: {type(target_)}')
            self._rotation_around_point = to_apysc_val_from_builtin.\
                get_copied_int_from_builtin_val(integer=rotation_around_point)
            self._rotation_around_point_diff = (
                self._rotation_around_point
                - self._before_rotation_around_point)
            self._set_basic_animation_settings(
                target=target,
                duration=duration,
                delay=delay,
                easing=easing)
            super(AnimationRotationAroundPoint, self).__init__(
                variable_name=variable_name)

    def _get_animation_func_expression(self) -> str:
        """
        Get a animation function expression.

        Returns
        -------
        expression : str
            Animation function expression.
        """
        from apysc._type import value_util
        diff_str: str = value_util.get_value_str_for_expression(
            value=self._rotation_around_point_diff)
        x_str: str = value_util.get_value_str_for_expression(value=self._x)
        y_str: str = value_util.get_value_str_for_expression(value=self._y)
        return f'\n  .rotate({diff_str}, {x_str}, {y_str});'

    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to be inserted into the complete event
        handler's head.

        Returns
        -------
        expression : str
            An expression to be inserted into the complete event
            handler's head.
        """
        from apysc._display import rotation_interface_helper
        from apysc._display.rotation_around_point_interface import \
            RotationAroundPointInterface
        from apysc._type import value_util
        expression: str = ''
        if isinstance(self._target, RotationAroundPointInterface):
            self._target.\
                _initialize_rotation_around_point_if_not_initialized()
            key_exp_str: str = rotation_interface_helper.\
                get_coordinates_key_for_expression(
                    x=self._x, y=self._y).value
            target_rotation_around_point_value_str: str = value_util.\
                get_value_str_for_expression(
                    value=self._target._rotation_around_point)
            rotation_around_point_value_str: str = value_util.\
                get_value_str_for_expression(
                    value=self._rotation_around_point)
            expression = (
                f'{target_rotation_around_point_value_str}'
                f'[{key_exp_str}] = '
                f'{rotation_around_point_value_str};'
            )
        return expression



"""Class implementation for the animation_fill_color interface.
"""

from typing import TypeVar
from typing import Union

from apysc._animation.animation_fill_color import AnimationFillColor
from apysc._animation.animation_interface_base import AnimationInterfaceBase
from apysc._animation.easing import Easing
from apysc._type.int import Int
from apysc._type.string import String

StrOrString = TypeVar('StrOrString', str, String)


class AnimationFillColorInterface(AnimationInterfaceBase):

    def animation_fill_color(
            self,
            fill_color: StrOrString,
            *,
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> AnimationFillColor:
        """
        Set the fill color animation setting.

        Notes
        -----
        To start this animation, you need to call the `start` method of
        the returned instance.

        Parameters
        ----------
        fill_color : str or String
            The final fill color (hex color code) of the animation.
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting.

        Returns
        -------
        animation_fill_color : AnimationFillColor
            Created animation setting instance.

        References
        ----------
        - animation_fill_color interface document
            - https://simon-ritchie.github.io/apysc/animation_fill_color.html
        - Animation interfaces duration setting document
            - https://simon-ritchie.github.io/apysc/animation_duration.html
        - Animation interfaces delay setting document
            - https://simon-ritchie.github.io/apysc/animation_delay.html
        - Each animation interface return value document
            - https://bit.ly/2XOoa8w
        - Sequential animation setting document
            - https://simon-ritchie.github.io/apysc/sequential_animation.html
        - animation_parallel interface document
            - https://simon-ritchie.github.io/apysc/animation_parallel.html
        - Easing enum document
            - https://simon-ritchie.github.io/apysc/easing_enum.html
        """
        animation_fill_color: AnimationFillColor = AnimationFillColor(
            target=self,
            fill_color=fill_color,
            duration=duration,
            delay=delay,
            easing=easing)
        return animation_fill_color



"""Each js assertion (console.assert) interface implementations.

Mainly following interfaces are defined:

- assert_equal
    JavaScript assertion interface for equal condition.
- assert_not_equal
    JavaScript assertion interface for not equal condition.
- assert_true
    JavaScript assertion interface for true condition.
- assert_false
    JavaScript assertion interface for false condition.
- assert_arrays_equal
    JavaScript assertion interface for Array values equal condition.
- assert_arrays_not_equal
    JavaScript assertion interface for Array values not equal condition.
- assert_dicts_equal
    JavaScript assertion interface for Dictionary values equal condition.
- assert_dicts_not_equal
    JavaScript assertion interface for Dictionary values not equal
    condition.
- assert_defined
    JavaScript assertion interface for defined (not undefined)
    value condition.
- assert_undefined
    JavaScript assertion interface for undefined value condition.
"""

from typing import Any
from typing import Tuple


def assert_equal(
        left: Any, right: Any, *, msg: str = '') -> None:
    """
    JavaScript assertion interface for equal condition.

    Notes
    -----
    - If specified values are type of Array (or list, etc),
        then assert_arrays_equal function will be called instead of
        this function.
    - If specified value are type of Dictionary (or dict, etc),
        then assert_dicts_equal function will be called instead of
        this function.

    Parameters
    ----------
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_equal, locals_=locals(),
            module_name=__name__):
        from apysc._string import string_util
        for value in (left, right):
            if _value_type_is_array(value=value):
                assert_arrays_equal(left=left, right=right, msg=msg)
                return
        for value in (left, right):
            if _value_type_is_dict(value=value):
                assert_dicts_equal(left=left, right=right, msg=msg)
                return

        _trace_info(
            interface_label='assert_equal', left=left, right=right)

        left_str, right_str = _get_left_and_right_strs(
            left=left, right=right)

        msg = string_util.escape_str(string=msg)
        expression: str = (
            f'console.assert({left_str} === {right_str}, "{msg}");'
        )
        ap.append_js_expression(expression=expression)


def assert_not_equal(
        left: Any, right: Any, *, msg: str = '') -> None:
    """
    JavaScript assertion interface for not equal condition.

    Notes
    -----
    - If specified values are type of Array (or list, etc),
        then assert_arrays_not_equal function will be called instead of
        this function.
    - If specified value are type of Dictionary (or dict, etc),
        then assert_dicts_not_equal function will be called instead of
        this function.

    Parameters
    ----------
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_not_equal, locals_=locals(),
            module_name=__name__):
        from apysc._string import string_util
        for value in (left, right):
            if _value_type_is_array(value=value):
                assert_arrays_not_equal(left=left, right=right, msg=msg)
                return
        for value in (left, right):
            if _value_type_is_dict(value=value):
                assert_dicts_not_equal(left=left, right=right, msg=msg)
                return

        _trace_info(
            interface_label='assert_not_equal',
            left=left,
            right=right)
        left_str, right_str = _get_left_and_right_strs(
            left=left, right=right)

        msg = string_util.escape_str(string=msg)
        expression: str = (
            f'console.assert({left_str} !== {right_str}, "{msg}");'
        )
        ap.append_js_expression(expression=expression)


def assert_true(
        value: Any, *, type_strict: bool = True, msg: str = '') -> None:
    """
    JavaScript assertion interface for true condition.

    Parameters
    ----------
    value : *
        Target value to check.
    type_strict : bool, default True
        Whether strictly check actual value or not.
        For example, if type_strict is True, interger 1 will
        fail, on the contrary (if type_strict is False), integer 1
        will pass test.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_true, locals_=locals(),
            module_name=__name__):
        from apysc._string import string_util
        _trace_info(
            interface_label='assert_true', left='true', right=value)
        _, value_str = _get_left_and_right_strs(
            left='_', right=value)

        msg = string_util.escape_str(string=msg)
        expression: str = (
            f'console.assert({value_str} =='
        )
        expression = _add_equal_if_type_strict_setting_is_true(
            expression=expression, type_strict=type_strict)
        expression += f' true, "{msg}");'
        ap.append_js_expression(expression=expression)


def assert_false(
        value: Any, *, type_strict: bool = True, msg: str = '') -> None:
    """
    JavaScript assertion interface for false condition.

    Parameters
    ----------
    value : *
        Target value to check.
    type_strict : bool, default True
        Whether strictly check actual value or not.
        For example, if type_strict is True, interger 0 will
        fail, on the contrary (if type_strict is False), integer 0
        will pass test.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_false, locals_=locals(),
            module_name=__name__):
        from apysc._string import string_util
        _trace_info(
            interface_label='assert_false', left='false', right=value)
        _, value_str = _get_left_and_right_strs(
            left='_', right=value)

        msg = string_util.escape_str(string=msg)
        expression: str = (
            f'console.assert({value_str} =='
        )
        expression = _add_equal_if_type_strict_setting_is_true(
            expression=expression, type_strict=type_strict)
        expression += f' false, "{msg}");'
        ap.append_js_expression(expression=expression)


def assert_arrays_equal(
        left: Any, right: Any, *, msg: str = '') -> None:
    """
    JavaScript assertion interface for Array values equal condition.

    Notes
    -----
    This is used instead of assert_equal for Array class
    comparison (JavaScript can not compare arrays directly, like
    a Python, for example, `[1, 2] === [1, 2]` will be false).

    Parameters
    ----------
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_arrays_equal, locals_=locals(),
            module_name=__name__):
        _trace_arrays_or_dicts_assertion_info(
            interface_label='assert_arrays_equal',
            left=left, right=right)

        expression: str = _make_arrays_or_dicts_comparison_expression(
            left=left, right=right, msg=msg, not_condition=False)
        ap.append_js_expression(expression=expression)


def assert_arrays_not_equal(
        left: Any, right: Any, *, msg: str = '') -> None:
    """
    JavaScript assertion interface for Array values not equal condition.

    Notes
    -----
    This is used instead of assert_not_equal for Array class
    comparison (JavaScript can not compare arrays directly, like
    a Python, for example, `[1, 2] === [1, 2]` will be false).

    Parameters
    ----------
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_arrays_not_equal, locals_=locals(),
            module_name=__name__):
        _trace_arrays_or_dicts_assertion_info(
            interface_label='assert_arrays_not_equal',
            left=left, right=right)

        expression: str = _make_arrays_or_dicts_comparison_expression(
            left=left, right=right, msg=msg, not_condition=True)
        ap.append_js_expression(expression=expression)


def assert_dicts_equal(left: Any, right: Any, *, msg: str = '') -> None:
    """
    JavaScript assertion interface for Dictionary values equal
    condition.

    Notes
    -----
    This is used instead of assert_equal for Dictionary class
    comparison (JavaScript can not compare dictionary (Object)
    directly, like a Python, for example, `{"a": 10} === {"a": 10}`
    will be false).

    Parameters
    ----------
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_dicts_equal, locals_=locals(),
            module_name=__name__):
        _trace_arrays_or_dicts_assertion_info(
            interface_label='assert_dicts_equal',
            left=left, right=right)

        expression: str = _make_arrays_or_dicts_comparison_expression(
            left=left, right=right, msg=msg, not_condition=False)
        ap.append_js_expression(expression=expression)


def assert_dicts_not_equal(
        left: Any, right: Any, *, msg: str = '') -> None:
    """
    JavaScript assertion interface for Dictionary values not equal
    condition.

    Notes
    -----
    This is used instead of assert_not_equal for Dictionary class
    comparison (JavaScript can not compare dictionary (Object)
    directly, like a Python, for example, `{"a": 10} !== {"a": 10}`
    will be true).

    Parameters
    ----------
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_dicts_not_equal, locals_=locals(),
            module_name=__name__):
        _trace_arrays_or_dicts_assertion_info(
            interface_label='assert_dicts_not_equal',
            left=left, right=right)

        expression: str = _make_arrays_or_dicts_comparison_expression(
            left=left, right=right, msg=msg, not_condition=True)
        ap.append_js_expression(expression=expression)


def assert_defined(value: Any, *, msg: str = '') -> None:
    """
    JavaScript assertion interface for defined (not undefined)
    value condition.

    Parameters
    ----------
    value : *
        Target value to check.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_defined, locals_=locals(),
            module_name=__name__):
        from apysc._string import string_util
        _trace_info(
            interface_label='assert_defined', left='other than undefined',
            right=value)
        _, value_str = _get_left_and_right_strs(
            left='_', right=value)

        msg = string_util.escape_str(string=msg)
        expression: str = (
            f'console.assert(!_.isUndefined({value_str}), "{msg}");'
        )
        ap.append_js_expression(expression=expression)


def assert_undefined(value: Any, *, msg: str = '') -> None:
    """
    JavaScript assertion interface for undefined value condition.

    Parameters
    ----------
    value : *
        Target value to check.
    msg : str, optional
        Message to display when assertion failed.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=assert_undefined, locals_=locals(),
            module_name=__name__):
        from apysc._string import string_util
        _trace_info(
            interface_label='assert_undefined', left='undefined',
            right=value)
        _, value_str = _get_left_and_right_strs(
            left='_', right=value)

        msg = string_util.escape_str(string=msg)
        expression: str = (
            f'console.assert(_.isUndefined({value_str}), "{msg}");'
        )
        ap.append_js_expression(expression=expression)


def _make_arrays_or_dicts_comparison_expression(
        *, left: Any, right: Any, msg: str,
        not_condition: bool) -> str:
    """
    Make arrays or dicts comparison (assert_arrays_equal,
    assert_arrays_not_equal, assert_dicts_equal, or
    assert_dicts_not_equal) expression string.

    Parameters
    ----------
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.
    msg : str, optional
        Message to display when assertion failed.
    not_condition : bool
        Boolean value whether this expression is not condition
        (assert_arrays_not_equal) or not.

    Returns
    -------
    expression : str
        Result expression string.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=_make_arrays_or_dicts_comparison_expression,
            locals_=locals(),
            module_name=__name__):
        from apysc._string import string_util
        from apysc._type import value_util
        left_exp_str: str = value_util.get_value_str_for_expression(
            value=left)
        right_exp_str: str = value_util.get_value_str_for_expression(
            value=right)
        msg = string_util.escape_str(string=msg)
        if not_condition:
            not_condition_str: str = '!'
        else:
            not_condition_str = ''
        expression: str = (
            f'console.assert({not_condition_str}_.isEqual({left_exp_str}, '
            f'{right_exp_str}), "{msg}");'
        )
        return expression


def _trace_arrays_or_dicts_assertion_info(
        *, interface_label: str, left: Any, right: Any) -> None:
    """
    Append arrays or dicts value's information trace expression.

    Parameters
    ----------
    interface_label : str
        Target assertion interface label, e.g., 'assert_arrays_equal'.
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=_trace_arrays_or_dicts_assertion_info, locals_=locals(),
            module_name=__name__):
        from apysc._type import value_util
        left_exp_str: str = value_util.get_value_str_for_expression(
            value=left)
        if isinstance(left, dict):
            left_exp_str = left_exp_str.replace('"', '')
        right_exp_str: str = value_util.get_value_str_for_expression(
            value=right)
        if isinstance(right, dict):
            right_exp_str = right_exp_str.replace('"', '')
        if isinstance(left, (ap.Array, ap.Dictionary)):
            value_str: str = value_util.get_value_str_for_expression(
                value=left.value)
            value_str = value_str.replace('"', '')
            left_info_str: str = f'{left_exp_str} ({value_str})'
        else:
            left_info_str = left_exp_str
        right_info_str = right_exp_str
        _trace_info(
            interface_label=interface_label,
            left=left_info_str,
            right=right_info_str)


def _value_type_is_array(*, value: Any) -> bool:
    """
    Get a boolean value whether the specified value is
    Array type or not.

    Parameters
    ----------
    value : *
        Target value to check.

    Returns
    -------
    result : bool
        If the value type is Array, True will be returned.
    """
    import apysc as ap
    if isinstance(value, ap.Array):
        return True
    return False


def _value_type_is_dict(*, value: Any) -> bool:
    """
    Get a boolean value whether the specified value is
    Dictionary type or not.

    Parameters
    ----------
    value : *
        Target value to check.

    Returns
    -------
    result : bool
        If the value type is Dictionary, True will be returned.
    """
    from apysc._type.dictionary_structure import DictionaryStructure
    if isinstance(value, DictionaryStructure):
        return True
    return False


def _add_equal_if_type_strict_setting_is_true(
        *, expression: str, type_strict: bool) -> str:
    """
    Add single equal character to expression if type_string setting
    is True.

    Parameters
    ----------
    expression : str
        Expression to be added.
    type_strict: bool
        Type strict setting value.

    Returns
    -------
    expression : str
        If type_string setting is true, then single equal character
        will be added to tail.
    """
    if not type_strict:
        return expression
    expression += '='
    return expression


def _get_left_and_right_strs(
        *, left: Any, right: Any) -> Tuple[str, str]:
    """
    Get left and right value strings from specified values.

    Parameters
    ----------
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.

    Returns
    -------
    left_str : str
        Left-side value's string. If value is string, this will be
        wrapped by double quotation.
    right_str : str
        Right-side value's string. If value is string, this will be
        wrapped by double quotation.
    """
    from apysc._type import value_util
    left_str: str = value_util.get_value_str_for_expression(
        value=left)
    right_str: str = value_util.get_value_str_for_expression(value=right)
    return left_str, right_str


def _trace_info(*, interface_label: str, left: Any, right: Any) -> None:
    """
    Append trace expression of specified values.

    Parameters
    ----------
    interface_label : str
        Target assertion interface label, e.g., 'assert_equal'.
    left : *
        Left-side value to compare.
    right : *
        Right-side value to compare.
    """
    import apysc as ap
    with ap.DebugInfo(
            callable_=_trace_info, locals_=locals(),
            module_name=__name__):
        from apysc._type.variable_name_interface import VariableNameInterface
        info: str = f'[{interface_label}]'
        if isinstance(left, VariableNameInterface):
            info += f'\nLeft-side variable name: {left.variable_name}'
        if isinstance(right, VariableNameInterface):
            info += f'\nRight-side variable name: {right.variable_name}'
        ap.trace(info, '\nLeft value:', left, 'right value:', right)



"""Class implementation for the ellipse height interface.
"""

from typing import Dict

from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class EllipseHeightInterface(
        VariableNameInterface, RevertInterface, AttrLinkingInterface):

    _ellipse_height: Int

    def _initialize_ellipse_height_if_not_initialized(self) -> None:
        """
        Initialize _ellipse_height attribute if it hasn't
        been initialized yet.
        """
        if hasattr(self, '_ellipse_height'):
            return
        self._ellipse_height = Int(0)

        self._append_ellipse_height_attr_linking_setting()

    def _append_ellipse_height_attr_linking_setting(self) -> None:
        """
        Append a ellipse-height attribute linking setting.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ellipse_height_attr_linking_setting,
                locals_=locals(),
                module_name=__name__, class_=EllipseHeightInterface):
            self._append_applying_new_attr_val_exp(
                new_attr=self._ellipse_height, attr_name='ellipse_height')
            self._append_attr_to_linking_stack(
                attr=self._ellipse_height, attr_name='ellipse_height')

    @property
    def ellipse_height(self) -> Int:
        """
        Get ellipse height value.

        Returns
        -------
        ellipse_height : Int
            Ellipse height value.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> rectangle.ellipse_width = ap.Int(10)
        >>> rectangle.ellipse_height = ap.Int(15)
        >>> rectangle.ellipse_height
        Int(15)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='ellipse_height', locals_=locals(),
                module_name=__name__, class_=EllipseHeightInterface):
            from apysc._type import value_util
            self._initialize_ellipse_height_if_not_initialized()
            return value_util.get_copy(value=self._ellipse_height)

    @ellipse_height.setter
    def ellipse_height(self, value: Int) -> None:
        """
        Update ellipse height value.

        Parameters
        ----------
        value : int or Int
            Ellipse height value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='ellipse_height', locals_=locals(),
                module_name=__name__, class_=EllipseHeightInterface):
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=value)
            if isinstance(value, int):
                value = ap.Int(value)
            self._ellipse_height = value
            self._ellipse_height.\
                _append_incremental_calc_substitution_expression()
            self._append_ellipse_height_update_expression()

            self._append_ellipse_height_attr_linking_setting()

    def _append_ellipse_height_update_expression(self) -> None:
        """
        Append ellipse height updating expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ellipse_height_update_expression,
                locals_=locals(),
                module_name=__name__, class_=EllipseHeightInterface):
            from apysc._type import value_util
            self._initialize_ellipse_height_if_not_initialized()
            if hasattr(self, '_ellipse_width'):
                width_value_str: str = value_util.get_value_str_for_expression(
                    value=getattr(self, '_ellipse_width'))
            else:
                width_value_str = value_util.get_value_str_for_expression(
                    value=0)
            height_value_str: str = value_util.get_value_str_for_expression(
                value=self._ellipse_height)
            expression: str = (
                f'{self.variable_name}.radius({width_value_str}, '
                f'{height_value_str});'
            )
            ap.append_js_expression(expression=expression)

    _ellipse_height_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make the value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_ellipse_height_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_ellipse_height_snapshots',
            value=int(self._ellipse_height._value),
            snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert the value if the snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._ellipse_height._value = self._ellipse_height_snapshots[
            snapshot_name]



"""Class implementation for dictionary.
"""

from typing import Any
from typing import Dict
from typing import Generic
from typing import TypeVar
from typing import Union

from apysc._event.custom_event_interface import CustomEventInterface
from apysc._type.any_value import AnyValue
from apysc._type.boolean import Boolean
from apysc._type.copy_interface import CopyInterface
from apysc._type.dictionary_structure import DictionaryStructure
from apysc._type.expression_string import ExpressionString
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.revert_interface import RevertInterface
from apysc._type.string import String
from apysc._type.variable_name_interface import VariableNameInterface

DefaultType = TypeVar('DefaultType')

_BuiltinKeys = Union[str, int, float]
_K = TypeVar('_K')
_V = TypeVar('_V')


class Dictionary(
        Generic[_K, _V],
        CopyInterface, RevertInterface, DictionaryStructure,
        CustomEventInterface):
    """
    Dictionary class for the apysc library.

    References
    ----------
    - Dictionary document
        - https://simon-ritchie.github.io/apysc/dictionary.html
    - Dictionary class generic type settings document
        - https://bit.ly/3HL8HaF
    """

    _initial_value: Union[Dict[_K, _V], 'Dictionary']
    _value: Dict[_K, _V]

    def __init__(self, value: Union[Dict[_K, _V], 'Dictionary']) -> None:
        """
        Dictionary class for the apysc library.

        Parameters
        ----------
        value : dict or Dictionary
            Initial dictionary value.

        References
        ----------
        - Dictionary document
            - https://simon-ritchie.github.io/apysc/dictionary.html
        - Dictionary class generic type settings document
            - https://bit.ly/3HL8HaF

        Examples
        --------
        >>> import apysc as ap
        >>> dictionary: ap.Dictionary = ap.Dictionary({'a': 10})
        >>> dictionary
        Dictionary({'a': 10})

        >>> dictionary['a']
        10

        >>> dictionary['b'] = 20
        >>> dictionary['b']
        20
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            with TemporaryNotHandlerScope():
                TYPE_NAME: str = var_names.DICTIONARY
                self._validate_acceptable_value_type(value=value)
                self._initial_value = value
                self._type_name = TYPE_NAME
                self._value = self._get_dict_value(value=value)
                self.variable_name = expression_variables_util.\
                    get_next_variable_name(type_name=TYPE_NAME)
                self._append_constructor_expression()

    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=self._initial_value)
            expression: str = f'var {self.variable_name} = {value_str};'
            ap.append_js_expression(expression=expression)

    def _get_dict_value(
            self, *,
            value: Union[Dict[_K, _V], 'Dictionary']) -> Dict[_K, _V]:
        """
        Get a dict value from specified value.

        Parameters
        ----------
        value : dict or Dictionary
            Specified dictionary value.

        Returns
        -------
        dict_val : dict
            Converted dict value.
        """
        if isinstance(value, Dictionary):
            return value._value
        return value

    def _validate_acceptable_value_type(
            self, *, value: Union[Dict[_K, _V], 'Dictionary']) -> None:
        """
        Validate that specified value is acceptable type or not.

        Parameters
        ----------
        value : dict or Dictionary
            Dictionary value to check.

        Raises
        ------
        TypeError
            If specified value's type is not dict or Dictionary.
        """
        if isinstance(value, (dict, Dictionary)):
            return
        raise TypeError(
            'Not acceptable value type is specified.'
            f'\nSpecified valkue type is: {type(value)}'
            '\nAcceptable types are: dict and Dictionary'
        )

    @property
    def value(self) -> Union[Dict[_K, _V], 'Dictionary']:
        """
        Get a current dict value.

        Returns
        -------
        value : dict
            Current dict value.

        References
        ----------
        - apysc basic data classes common value interface
            - https://bit.ly/3Be1aij

        Examples
        --------
        >>> import apysc as ap
        >>> dictionary: ap.Dictionary = ap.Dictionary({})
        >>> dictionary.value = {'a': 10}
        >>> dictionary.value
        {'a': 10}
        """
        return self._value

    @value.setter
    def value(self, value: Union[Dict[_K, _V], 'Dictionary']) -> None:
        """
        Set dictionary value.

        Parameters
        ----------
        value : dict or Dictionary.
            Dictionary value to set.

        References
        ----------
        apysc basic data classes common value interface
            https://bit.ly/3Be1aij
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='value', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            self._validate_acceptable_value_type(value=value)
            self._value = self._get_dict_value(value=value)
            self._append_value_setter_expression(value=value)

    def _append_value_setter_expression(
            self, *, value: Union[Dict[_K, _V], 'Dictionary']) -> None:
        """
        Append value's setter expression.

        Parameters
        ----------
        value : dict or Dictionary.
            Dictionary value to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_value_setter_expression,
                locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression: str = f'{self.variable_name} = {value_str};'
            ap.append_js_expression(expression=expression)

    _value_snapshot: Dict[str, Dict[_K, _V]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make values' snapthot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_value_snapshot',
            value={**self._value}, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert values if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._value = self._value_snapshot[snapshot_name]

    def __str__(self) -> str:
        """
        String conversion method.

        Returns
        -------
        string : str
            Converted value string.
        """
        if not hasattr(self, '_value'):
            return '{}'
        return str(self._value)

    def __repr__(self) -> str:
        """
        Get a representation string of this instance.

        Returns
        -------
        repr_str : str
            Representation string of this instance.
        """
        if not hasattr(self, '_value'):
            repr_str: str = 'Dictionary({})'
        else:
            repr_str = f'Dictionary({self._value})'
        return repr_str

    @property
    def length(self) -> Int:
        """
        Get length of this dictionary values.

        Returns
        -------
        length : Int
            This dictionary value's length.

        References
        ----------
        - Dictionary length interface document
            - https://simon-ritchie.github.io/apysc/dictionary_length.html

        Examples
        --------
        >>> import apysc as ap
        >>> dictionary: ap.Dictionary = ap.Dictionary({'a': 1, 'b': 2})
        >>> dictionary.length
        Int(2)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='length', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            length: ap.Int = ap.Int(len(self._value))
            self._append_length_expression(length=length)
            return length

    def _append_length_expression(self, *, length: Int) -> None:
        """
        Append length method expression.

        Parameters
        ----------
        length : Int
            Created length Int variable.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_length_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            expression: str = (
                f'{length.variable_name} = '
                f'Object.keys({self.variable_name}).length;'
            )
            ap.append_js_expression(expression=expression)

    def __len__(self) -> None:
        """
        This method is disabled and can't use from Dictionary instance.
        """
        raise Exception(
            'Dictionary instance can\'t apply len function.'
            ' Please use length property instead.')

    def __getitem__(self, key: Union[_K, ExpressionString]) -> _V:
        """
        Get a specified key's single value.

        Parameters
        ----------
        key : _K
            Dictionary key.

        Returns
        -------
        value : *
            Specified key's value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__getitem__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            self._validate_key_type_is_str_or_numeric(key=key)
            key_: _BuiltinKeys = self._get_builtin_type_key(key=key)
            has_key: bool = key_ in self._value
            if has_key:
                value: Any = self._value[key_]  # type: ignore
            else:
                value = ap.AnyValue(None)
            self._append_getitem_expression(key=key, value=value)
            return value

    def _get_builtin_type_key(
            self, *, key: Union[_K, ExpressionString]) -> _BuiltinKeys:
        """
        Get a built-in type's key (str, int, or float) from
        a specified key.

        Parameters
        ----------
        key : _K
            Target key value (including String, Int, and Number).

        Returns
        -------
        key : str or int or float
            Built-in type's key.
        """
        if isinstance(key, ExpressionString):
            return key.value
        if isinstance(key, (String, Int, Number)):
            return key._value
        key_: _BuiltinKeys = key  # type: ignore
        return key_

    def _append_getitem_expression(
            self, *, key: Union[_K, ExpressionString], value: Any) -> None:
        """
        Append __getitem__ expression.

        Parameters
        ----------
        key : _K
            Dictionary key.
        value : *
            Specified key's value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_getitem_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            if not isinstance(value, VariableNameInterface):
                value = ap.AnyValue(None)
            key_str: str = value_util.get_value_str_for_expression(value=key)
            expression: str = (
                f'var {value.variable_name} = '
                f'{self.variable_name}[{key_str}];'
            )
            ap.append_js_expression(expression=expression)

    def _validate_key_type_is_str_or_numeric(
            self, *, key: Union[_K, ExpressionString]) -> None:
        """
        Validate whether key value type is acceptable (str or int or
        flaot) or not.

        Parameters
        ----------
        key : _K
            Dictionary key to check.

        Raises
        ------
        ValueError
            If key type is not str, String, int, Int, float, or Number.
        """
        if isinstance(
                key,
                (str, String, int, Int, float, Number,
                 ExpressionString)):
            return
        raise ValueError(
            f'Unsupported key type is specified: {type(key)}, {key}'
            f'\nSuppoting types are: str, String, int, Int')

    def __setitem__(
            self, key: Union[_K, ExpressionString], value: _V) -> None:
        """
        Set value to a specified key position.

        Parameters
        ----------
        key : _K
            Dictionary key to set value.
        value : *
            Any value to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__setitem__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            key_: _BuiltinKeys = self._get_builtin_type_key(key=key)
            self._value[key_] = value  # type: ignore
            self._append_setitem_expression(key=key, value=value)

    def _append_setitem_expression(
            self, *, key: Union[_K, ExpressionString], value: _V) -> None:
        """
        Append __setitem__ method expression.

        Parameters
        ----------
        key : _K
            Dictionary key to check.
        value : *
            Any value to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_setitem_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            key_str: str = value_util.get_value_str_for_expression(value=key)
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression: str = (
                f'{self.variable_name}[{key_str}] = {value_str};'
            )
            ap.append_js_expression(expression=expression)

    def __delitem__(self, key: Union[_K, ExpressionString]) -> None:
        """
        Delete specified key's value from dictionary.

        Parameters
        ----------
        key : _K
            Dictionary key to delete.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__delitem__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            key_: _BuiltinKeys = self._get_builtin_type_key(key=key)
            if key_ in self._value:
                del self._value[key_]  # type: ignore
            self._append_delitem_expression(key=key)

    def _append_delitem_expression(
            self, *, key: Union[_K, ExpressionString]) -> None:
        """
        Append __delitem__ method expression.

        Parameters
        ----------
        key : _K
            Dictionary key to delete.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_delitem_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            from apysc._type import value_util
            key_str: str = value_util.get_value_str_for_expression(value=key)
            expression: str = (
                f'delete {self.variable_name}[{key_str}];'
            )
            ap.append_js_expression(expression=expression)

    def __eq__(self, other: Any) -> Any:
        """
        Equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. dict or Dictionary types are acceptable.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__eq__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            if isinstance(other, Dictionary):
                result: ap.Boolean = ap.Boolean(self._value == other._value)
            else:
                result = ap.Boolean(self._value == other)
                if isinstance(other, dict):
                    other = Dictionary(other)
            if isinstance(other, VariableNameInterface):
                self._append_eq_expression(result=result, other=other)
            return result

    def _append_eq_expression(
            self, *, result: Boolean,
            other: VariableNameInterface) -> None:
        """
        Append an __eq__ expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : Dictionary
            Dictionary other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_eq_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            expression: str = (
                f'{result.variable_name} = '
                f'_.isEqual({self.variable_name}, {other.variable_name});'
            )
            ap.append_js_expression(expression=expression)

    def __ne__(self, other: Any) -> Any:
        """
        Noe equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. dict or Dictionary types are acceptable.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__ne__', locals_=locals(),
                module_name=__name__, class_=Dictionary):
            if isinstance(other, dict):
                other = Dictionary(other)
            result: ap.Boolean = self == other
            result = result.not_
            if isinstance(other, VariableNameInterface):
                self._append_ne_expression(result=result, other=other)
            return result

    def _append_ne_expression(
            self, *, result: Boolean,
            other: VariableNameInterface) -> None:
        """
        Append a __ne__ expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : Dictionary
            Dictionary other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ne_expression, locals_=locals(),
                module_name=__name__, class_=Dictionary):
            expression: str = (
                f'{result.variable_name} = '
                f'!_.isEqual({self.variable_name}, {other.variable_name});'
            )
            ap.append_js_expression(expression=expression)

    def get(
            self, key: Union[_K, ExpressionString],
            *,
            default: DefaultType = None) -> DefaultType:
        """
        Get a specified key dictionary value. If this dictionary
        hasn't a specified key, then a default value will be returned.

        Parameters
        ----------
        key : _K
            Target key.
        default : DefaultType or None, optional
            Any default value.

        Returns
        -------
        result_value : Any
            Extracted value or a default value.

        References
        ----------
        - Dictionary get interface document
            - https://bit.ly/3FbSHNg

        Examples
        --------
        >>> from typing import Optional
        >>> import apysc as ap
        >>> dictionary: ap.Dictionary = ap.Dictionary({'a': 10})
        >>> value_1: Optional[int] = dictionary.get('a')
        >>> value_1
        10

        >>> value_2: Optional[int] = dictionary.get('b')
        >>> print(value_2)
        None

        >>> value_3: int = dictionary.get('c', default=0)
        >>> value_3
        0
        """
        key_: _BuiltinKeys = self._get_builtin_type_key(key=key)
        if key_ in self._value:
            result_value: Any = self._value[key_]  # type: ignore
        else:
            result_value = default
        self._append_get_expression(
            key=key, result_value=result_value, default=default)
        return result_value

    def _append_get_expression(
            self, *, key: Union[_K, ExpressionString],
            result_value: Any,
            default: Any) -> None:
        """
        Append the `get` method expression.

        Parameters
        ----------
        key : _K
            Target key.
        result_value : Any
            Extracted value or a default value.
        default : Any
            Any default value. Basically apysc types (e.g., Int, Number,
            String, and so on) are necessary.
        """
        import apysc as ap
        from apysc._type import value_util
        key_str: str = value_util.get_value_str_for_expression(value=key)
        if not isinstance(result_value, VariableNameInterface):
            result_value = AnyValue(value=None)
        result_value_str: str = value_util.get_value_str_for_expression(
            value=result_value)
        default_value_str: str = value_util.get_value_str_for_expression(
            value=default)
        expression: str = (
            f'if ({key_str} in {self.variable_name}) {{'
            f'\n  {result_value_str} = {self.variable_name}[{key_str}];'
            '\n}else {'
            f'\n  {result_value_str} = {default_value_str};'
            '\n}'
        )
        ap.append_js_expression(expression=expression)



"""The implementation of manipulating HTL and js expression files.

Mainly following interfaces are defined:

- empty_expression : Empty the current js expression data.
- append_js_expression : Append js expression.
- get_current_expression : Get current expression string.
- get_current_event_handler_scope_expression : Get a current
    event handler scope's expression string.
- exec_query : Execute a SQLite sql query.
"""

import sqlite3
from enum import Enum
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar


class TableName(Enum):
    NOT_EXISTING = 'not_existing'
    EXPRESSION_NORMAL = 'expression_normal'
    EXPRESSION_HANDLER = 'expression_handler'
    INDENT_NUM_NORMAL = 'indent_num_normal'
    INDENT_NUM_HANDLER = 'indent_num_handler'
    LAST_SCOPE = 'last_scope'
    EVENT_HANDLER_SCOPE_COUNT = 'event_handler_scope_count'
    LOOP_COUNT = 'loop_count'
    DEBUG_MODE_SETTING = 'debug_mode_setting'
    DEBUG_MODE_CALLABLE_COUNT = 'debug_mode_callable_count'
    STAGE_ELEM_ID = 'stage_elem_id'
    VARIABLE_NAME_COUNT = 'variable_name_count'
    HANDLER_CALLING_STACK = 'handler_calling_stack'
    CIRCULAR_CALLING_HANDLER_NAME = 'circular_calling_handler_name'
    STAGE_ID = 'stage_id'


_SQLITE_IN_MEMORY_SETTING: str = 'file::memory:?cache=shared'
connection = sqlite3.connect(_SQLITE_IN_MEMORY_SETTING, uri=True)
cursor = connection.cursor()

_C = TypeVar('_C', bound=Callable)


def _check_connection(func: _C) -> _C:
    """
    The decorator function to check a SQLite connection when a
    specified function calling, and if failed, create a new
    connection and recall a function.

    Parameters
    ----------
    func : Callable
        Target function to decorate.

    Returns
    -------
    new_func : Callable
        Decorated function.
    """

    def new_func(*args: Any, **kwargs: Any) -> Any:
        """
        Function for the decoration.

        Parameters
        ----------
        *args : list
            Any positional arguments.
        **kwargs : dict
            Any keyword arguments.

        Returns
        -------
        result : Any
            Any returned value.
        """
        global connection, cursor
        try:
            result: Any = func(*args, **kwargs)
        except Exception:
            connection = sqlite3.connect(_SQLITE_IN_MEMORY_SETTING, uri=True)
            cursor = connection.cursor()
            result = func(*args, **kwargs)
        return result

    return new_func  # type: ignore


@_check_connection
def _table_exists(*, table_name: TableName) -> bool:
    """
    Get a boolean value whether a specified table exists or not.

    Parameters
    ----------
    table_name : TableName
        Target table name.

    Returns
    -------
    result : bool
        If exists, returns True.
    """
    query: str = (
        'SELECT name FROM sqlite_master WHERE type = "table" '
        f'AND name = "{table_name.value}" LIMIT 1;'
    )
    cursor.execute(query)
    result: Optional[Tuple] = cursor.fetchone()
    connection.commit()
    if result:
        return True
    return False


def _make_create_table_query(
        *, table_name: TableName,
        column_ddl: str) -> str:
    """
    Make a create table sql query.

    Parameters
    ----------
    table_name : str
        Target table name.
    column_ddl : str
        Target table columns DDL string.
        e.g., '  id INTEGER, ...'

    Returns
    -------
    query : str
        A create table sql query.
    """
    query: str = (
        'CREATE TABLE IF NOT EXISTS '
        f'{table_name.value} ('
        f'\n{column_ddl}'
        '\n);'
    )
    return query


_EXPRESSION_TABLE_COLUMN_DDL: str = (
    '  id INTEGER PRIMARY KEY AUTOINCREMENT,'
    '\n  txt TEXT NOT NULL'
)


@_check_connection
def _create_expression_normal_table() -> None:
    """
    Create the normal expression data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.EXPRESSION_NORMAL,
        column_ddl=_EXPRESSION_TABLE_COLUMN_DDL)
    cursor.execute(query)


@_check_connection
def _create_expression_handler_table() -> None:
    """
    Create the handler expression data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.EXPRESSION_HANDLER,
        column_ddl=_EXPRESSION_TABLE_COLUMN_DDL)
    cursor.execute(query)


_INDENT_NUM_TABLE_COLUMN_DDL: str = '  num INTEGER NOT NULL'


@_check_connection
def _create_indent_num_normal_table() -> None:
    """
    Create the normal indentation number data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.INDENT_NUM_NORMAL,
        column_ddl=_INDENT_NUM_TABLE_COLUMN_DDL)
    cursor.execute(query)


@_check_connection
def _create_indent_num_handler_table() -> None:
    """
    Create the handler indentation number data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.INDENT_NUM_HANDLER,
        column_ddl=_INDENT_NUM_TABLE_COLUMN_DDL)
    cursor.execute(query)


@_check_connection
def _create_last_scope_table() -> None:
    """
    Create the last scope data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.LAST_SCOPE,
        column_ddl=(
            '  last_scope INTEGER NOT NULL'
        ))
    cursor.execute(query)


@_check_connection
def _create_event_handler_scope_count_table() -> None:
    """
    Create the event handler scope count value SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.EVENT_HANDLER_SCOPE_COUNT,
        column_ddl=(
            '  count INTEGER NOT NULL'
        ))
    cursor.execute(query)


@_check_connection
def _create_loop_count_table() -> None:
    """
    Create the loop count value SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.LOOP_COUNT,
        column_ddl=(
            '  count INTEGER NOT NULL'
        ))
    cursor.execute(query)


@_check_connection
def _create_debug_mode_setting_table() -> None:
    """
    Create the debug mode setting SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.DEBUG_MODE_SETTING,
        column_ddl=(
            '  is_debug_mode INTEGER NOT NULL'
        ))
    cursor.execute(query)


@_check_connection
def _create_debug_mode_callable_count_table() -> None:
    """
    Create the debug mode callable count data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.DEBUG_MODE_CALLABLE_COUNT,
        column_ddl=(
            '  id INTEGER PRIMARY KEY AUTOINCREMENT,'
            '\n  name TEXT NOT NULL,'
            '\n  count INTEGER NOT NULL'
        ))
    cursor.execute(query)


@_check_connection
def _create_stage_elem_id_table() -> None:
    """
    Create the stage element id data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.STAGE_ELEM_ID,
        column_ddl='  elem_id TEXT NOT NULL')
    cursor.execute(query)


@_check_connection
def _create_variable_name_count_table() -> None:
    """
    Create the variable name count data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.VARIABLE_NAME_COUNT,
        column_ddl=(
            '  id INTEGER PRIMARY KEY AUTOINCREMENT,'
            '\n  type_name TEXT NOT NULL,'
            '\n  count INTEGER NOT NULL'
        ))
    cursor.execute(query)


@_check_connection
def _create_handler_calling_stack_table() -> None:
    """
    Create the handler calling stack data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.HANDLER_CALLING_STACK,
        column_ddl=(
            '  id INTEGER PRIMARY KEY AUTOINCREMENT,'
            '\n  handler_name TEXT NOT NULL,'
            '\n  scope_count INTEGER NOT NULL,'
            '\n  variable_name TEXT NOT NULL'
        ))
    cursor.execute(query)


@_check_connection
def _create_circular_calling_handler_name_table() -> None:
    """
    Create the circular calling handler names data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.CIRCULAR_CALLING_HANDLER_NAME,
        column_ddl=(
            '  id INTEGER PRIMARY KEY AUTOINCREMENT,'
            '\n  handler_name TEXT NOT NULL,'
            '\n  prev_handler_name TEXT NOT NULL,'
            '\n  prev_variable_name TEXT NOT NULL'
        ))
    cursor.execute(query)


@_check_connection
def _create_stage_id_table() -> None:
    """
    Create the stage id data SQLite table.
    """
    query: str = _make_create_table_query(
        table_name=TableName.STAGE_ID,
        column_ddl='  stage_id INTEGER NOT NULL')
    cursor.execute(query)


def initialize_sqlite_tables_if_not_initialized() -> bool:
    """
    Initialize the sqlite tables if they have not been
    initialized yet.

    Returns
    -------
    initialized : bool
        If initialized, returns True.
    """
    table_exists: bool = _table_exists(
        table_name=TableName.EXPRESSION_NORMAL)
    if table_exists:
        return False
    _create_expression_normal_table()
    _create_expression_handler_table()
    _create_indent_num_normal_table()
    _create_indent_num_handler_table()
    _create_last_scope_table()
    _create_event_handler_scope_count_table()
    _create_loop_count_table()
    _create_debug_mode_setting_table()
    _create_debug_mode_callable_count_table()
    _create_stage_elem_id_table()
    _create_variable_name_count_table()
    _create_handler_calling_stack_table()
    _create_circular_calling_handler_name_table()
    _create_stage_id_table()
    return True


def empty_expression() -> None:
    """
    Empty the current js expression data.
    """
    initialize_sqlite_tables_if_not_initialized()
    for table_name in TableName:
        if table_name == TableName.NOT_EXISTING:
            continue
        query: str = f'DELETE FROM {table_name.value};'
        cursor.execute(query)
    connection.commit()


def append_js_expression(*, expression: str) -> None:
    """
    Append js expression.

    Parameters
    ----------
    expression : str
        JavaScript Expression string.
    """
    from apysc._expression import indent_num
    from apysc._expression import last_scope
    from apysc._string import indent_util
    initialize_sqlite_tables_if_not_initialized()
    current_indent_num: int = indent_num.get_current_indent_num()
    expression = indent_util.append_spaces_to_expression(
        expression=expression, indent_num=current_indent_num)
    expression = expression.replace('"', '""')
    table_name: TableName = _get_expression_table_name()
    query: str = (
        f'INSERT INTO {table_name.value}(txt) '
        f'VALUES ("{expression}");'
    )
    cursor.execute(query)
    connection.commit()
    last_scope.set_last_scope(value=last_scope.LastScope.NORMAL)


def _get_expression_table_name() -> TableName:
    """
    Get a expression table name. This value will be switched whether
    current scope is event handler's one or not.

    Returns
    -------
    table_name : str
        Target expression table name.
    """
    from apysc._expression import event_handler_scope
    event_handler_scope_count: int = \
        event_handler_scope.get_current_event_handler_scope_count()
    if event_handler_scope_count == 0:
        return TableName.EXPRESSION_NORMAL
    return TableName.EXPRESSION_HANDLER


def get_current_expression() -> str:
    """
    Get a current expression's string.

    Notes
    -----
    If it is necessary to get event handler scope's expression,
    then use get_current_event_handler_scope_expression function
    instead.

    Returns
    -------
    current_expression : str
        Current expression's string.
    """
    current_expression: str = _get_current_expression(
        table_name=TableName.EXPRESSION_NORMAL)
    return current_expression


def get_current_event_handler_scope_expression() -> str:
    """
    Get a current event handler scope's expression string.

    Notes
    -----
    If it is necessary to get normal scope's expression, then use
    get_current_expression function instead.

    Returns
    -------
    current_expression : str
        Current expression's string.
    """
    current_expression: str = _get_current_expression(
        table_name=TableName.EXPRESSION_HANDLER)
    return current_expression


def _get_current_expression(*, table_name: TableName) -> str:
    """
    Get a current expression string from a specified table.

    Parameters
    ----------
    table_name : TableName
        Target table name.

    Returns
    -------
    current_expression : str
        Current expression string.
    """
    initialize_sqlite_tables_if_not_initialized()
    query: str = (
        f'SELECT txt FROM {table_name.value}')
    cursor.execute(query)
    result: List[Tuple[str]] = cursor.fetchall()
    if not result:
        return ''
    expressions: List[str] = [tpl[0] for tpl in result]
    current_expression = '\n'.join(expressions)
    return current_expression


class _LimitClauseCantUseError(Exception):
    ...


def _validate_limit_clause(*, sql: str) -> None:
    """
    Validate whether a LIMIT clause is used in a UPDATE or DELETE sql.

    Parameters
    ----------
    sql : str
        Target sql.

    Raises
    ------
    _LimitClauseCantUseError
        If the LIMIT clause used in a DELETE or UPDATE sql.
    """
    sql_: str = sql.lower()
    if 'delete ' not in sql_ and 'update ' not in sql_:
        return
    if 'limit ' not in sql_:
        return
    raise _LimitClauseCantUseError(
        f'LIMIT clause cannot use in the UPDATE or DELETE sql: {sql_}')


def exec_query(*, sql: str, commit: bool = True) -> None:
    """
    Execute a SQLite sql query.

    Parameters
    ----------
    sql : str
        Target sql.
    commit : bool, default True
        A boolean value whether commit the transaction after the
        sql query or not.

    Raises
    ------
    _LimitClauseCantUseError
        If the LIMIT clause used in a DELETE or UPDATE sql.
    """
    _validate_limit_clause(sql=sql)
    initialize_sqlite_tables_if_not_initialized()
    cursor.execute(sql)
    if commit:
        connection.commit()



"""2-dimensional geometry point class implementation.
"""

from typing import Any
from typing import Dict
from typing import Union

from apysc._event.custom_event_interface import CustomEventInterface
from apysc._type.dictionary_structure import DictionaryStructure
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface

_Int = Union[int, Int]


class Point2D(
        VariableNameInterface, RevertInterface, DictionaryStructure,
        CustomEventInterface):
    """
    2-dimensional geometry point class.
    """

    _x: Int
    _y: Int

    def __init__(self, x: _Int, y: _Int) -> None:
        """
        2-dimensional geometry point.

        Parameters
        ----------
        x : int or Int
            X-coordinate.
        y : int or Int
            Y-coordinate.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Point2D):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._validation.number_validation import validate_integer
            validate_integer(integer=x)
            validate_integer(integer=y)
            if isinstance(x, int):
                x = ap.Int(x)
            if isinstance(y, int):
                y = ap.Int(y)
            self._x = x
            self._y = y
            self.variable_name = \
                expression_variables_util.get_next_variable_name(
                    type_name=var_names.POINT2D)
            self._append_constructor_expression()

    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Point2D):
            expression: str = (
                f'var {self.variable_name} = {{'
                f'"x": {self._x.variable_name}, '
                f'"y": {self._y.variable_name}}};')
            ap.append_js_expression(expression=expression)

    @property
    def x(self) -> Int:
        """
        X-coordinate property.

        Returns
        -------
        x : Int
            X-coordinate.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='x', locals_=locals(),
                module_name=__name__, class_=Point2D):
            x: ap.Int = ap.Int(self._x._value)
            self._append_x_getter_expression(x=x)
            return x

    @x.setter
    def x(self, value: Int) -> None:
        """
        Update x-coordinate property.

        Parameters
        ----------
        value : Int
            X-coordinate to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='x', locals_=locals(),
                module_name=__name__, class_=Point2D):
            from apysc._validation.number_validation import validate_integer
            validate_integer(integer=value)
            if isinstance(value, int):
                value = ap.Int(value)
            self._x = value
            self._x._append_incremental_calc_substitution_expression()
            self._append_x_setter_expression(value=value)

    def _append_x_getter_expression(self, *, x: Int) -> None:
        """
        Append x property getter expression.

        Parameters
        ----------
        x : Int
            Target x value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_x_getter_expression, locals_=locals(),
                module_name=__name__, class_=Point2D):
            expression: str = (
                f'{x.variable_name} = '
                f'{self.variable_name}["x"];'
            )
            ap.append_js_expression(expression=expression)

    def _append_x_setter_expression(self, *, value: Int) -> None:
        """
        Append x property setter expression.

        Parameters
        ----------
        value : Int
            X-coordinate to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_x_setter_expression, locals_=locals(),
                module_name=__name__, class_=Point2D):
            expression: str = (
                f'{self.variable_name}["x"] = {value.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    @property
    def y(self) -> Int:
        """
        Y-coordinate property.

        Returns
        -------
        y : Int
            Y-coordinate.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='y', locals_=locals(),
                module_name=__name__, class_=Point2D):
            y: ap.Int = ap.Int(self._y._value)
            self._append_y_getter_expression(y=y)
            return y

    @y.setter
    def y(self, value: Int) -> None:
        """
        Update y-coordinate property.

        Parameters
        ----------
        value : Int
            Y-coordinate to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='y', locals_=locals(),
                module_name=__name__, class_=Point2D):
            from apysc._validation.number_validation import validate_integer
            validate_integer(integer=value)
            if isinstance(value, int):
                value = ap.Int(value)
            self._y = value
            self._y._append_incremental_calc_substitution_expression()
            self._append_y_setter_expression(value=value)

    def _append_y_getter_expression(self, *, y: Int) -> None:
        """
        Append y property getter expression.

        Parameters
        ----------
        y : Int
            Target y value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_y_getter_expression, locals_=locals(),
                module_name=__name__, class_=Point2D):
            expression: str = (
                f'{y.variable_name} = '
                f'{self.variable_name}["y"];'
            )
            ap.append_js_expression(expression=expression)

    def _append_y_setter_expression(self, *, value: Int) -> None:
        """
        Append y property setter expression.

        Parameters
        ----------
        value : Int
            Y-coordinate to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_y_setter_expression,
                locals_=locals(),
                module_name=__name__, class_=Point2D):
            expression: str = (
                f'{self.variable_name}["y"] = {value.variable_name};'
            )
            ap.append_js_expression(expression=expression)

    def __eq__(self, other: Any) -> Any:
        """
        Equal comparison method.

        Parameters
        ----------
        other : Any
            Other value to compare.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__eq__', locals_=locals(),
                module_name=__name__, class_=Point2D):
            result: ap.Boolean
            if not isinstance(other, Point2D):
                result = ap.Boolean(False)
                return result
            if other.x == self.x and other.y == self.y:
                result = ap.Boolean(True)
                return result
            result = ap.Boolean(False)
            return result

    def __ne__(self, other: Any) -> Any:
        """
        Not equal comparison method.

        Parameters
        ----------
        other : Any
            Other value to compare.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__ne__', locals_=locals(),
                module_name=__name__, class_=Point2D):
            result: ap.Boolean = self == other
            result = result.not_
            return result

    _x_snapshots: Dict[str, int]
    _y_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make values' snapshots.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_x_snapshots',
            value=int(self._x._value), snapshot_name=snapshot_name)
        self._set_single_snapshot_val_to_dict(
            dict_name='_y_snapshots',
            value=int(self._y._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert values if snapshots exist.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._x._value = self._x_snapshots[snapshot_name]
        self._y._value = self._y_snapshots[snapshot_name]



"""Implementations of the Path class.
"""

from typing import List

from apysc._display import graphics
from apysc._display.line_base import LineBase
from apysc._geom.path_data_base import PathDataBase


class Path(LineBase):
    """
    The path vector graphics class.
    """

    _path_data_list: List[PathDataBase]

    def __init__(
            self,
            *,
            parent: 'graphics.Graphics',
            path_data_list: List[PathDataBase]) -> None:
        """
        Create a path vector graphics.

        Parameters
        ----------
        parent : Graphics
            Graphics instance to link this graphic.
        path_data_list : list of PathDataBase
            Target path data settings, such as the ap.PathData.MoveTo.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Path):
            from apysc._display.graphics import Graphics
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._validation import path_validation
            path_validation.validate_path_data_list(
                path_data_list=path_data_list)
            parent_graphics: Graphics = parent
            variable_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.PATH)
            super(Path, self).__init__(
                parent=parent, x=0, y=0, variable_name=variable_name)
            self._path_data_list = path_data_list
            self._set_initial_basic_values(parent=parent)
            self._append_constructor_expression()
            self._set_line_setting_if_not_none_value_exists(
                parent_graphics=parent_graphics)
            self._set_overflow_visible_setting()

    def _append_constructor_expression(self) -> None:
        """
        Append a constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Path):
            from apysc._display.stage import get_stage_variable_name
            from apysc._geom.path_data_util import \
                make_paths_expression_from_list
            from apysc._string import indent_util
            stage_variable_name: str = get_stage_variable_name()
            path_data_expression: str = make_paths_expression_from_list(
                path_data_list=self._path_data_list)
            INDENT_NUM: int = 2
            expression: str = (
                f'var {self.variable_name} = {stage_variable_name}'
                f'\n  .path({path_data_expression})'
                '\n  .attr({'
            )
            expression = self._append_basic_vals_expression(
                expression=expression,
                indent_num=INDENT_NUM)
            spaces: str = indent_util.make_spaces_for_html(
                indent_num=INDENT_NUM)
            if self._fill_color._value == '':
                expression += f'\n{spaces}fill: "transparent",'
            expression += '\n  });'
            ap.append_js_expression(expression=expression)

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Path('<variable_name>')`).
        """
        repr_str: str = f"Path('{self.variable_name}')"
        return repr_str



"""Implementations of the Rectangle class.
"""

from typing import Union

from apysc._display import graphics
from apysc._display.ellipse_height_interface import EllipseHeightInterface
from apysc._display.ellipse_width_interface import EllipseWidthInterface
from apysc._display.height_interface import HeightInterface
from apysc._display.line_base import LineBase
from apysc._display.width_interface import WidthInterface
from apysc._type.int import Int


class Rectangle(
        LineBase, WidthInterface, HeightInterface,
        EllipseWidthInterface, EllipseHeightInterface):
    """
    The rectangle vector graphics class.

    References
    ----------
    - Graphics draw_rect interface document
        - https://simon-ritchie.github.io/apysc/graphics_draw_rect.html
    """

    def __init__(
            self, *, parent: 'graphics.Graphics',
            x: Union[int, Int],
            y: Union[int, Int],
            width: Union[int, Int],
            height: Union[int, Int]) -> None:
        """
        Create a rectangle vector graphics.

        Parameters
        ----------
        parent : Graphics
            Graphics instance to link this graphics.
        x : int or Int
            X-coordinate to start drawing.
        y : int or Int
            Y-coordinate to start drawing.
        width : int or Int
            Rectangle width.
        height : int or Int
            Rectangle height.

        References
        ----------
        - Graphics draw_rect interface document
            - https://simon-ritchie.github.io/apysc/graphics_draw_rect.html
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Rectangle):
            from apysc._display.graphics import Graphics
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._validation import size_validation
            parent_graphics: Graphics = parent
            variable_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.RECTANGLE)
            super(Rectangle, self).__init__(
                parent=parent, x=x, y=y, variable_name=variable_name)
            size_validation.validate_size_is_gte_zero(size=width)
            size_validation.validate_size_is_gte_zero(size=height)
            self._update_width_and_skip_appending_exp(value=width)
            self._update_height_and_skip_appending_exp(value=height)
            self._set_initial_basic_values(parent=parent)
            self._append_constructor_expression()
            self._set_line_setting_if_not_none_value_exists(
                parent_graphics=parent_graphics)
            self._set_overflow_visible_setting()

    def __repr__(self) -> str:
        """
        Get a string representation of this instance (for the sake of
        debugging).

        Returns
        -------
        repr_str : str
            Type name and variable name will be set
            (e.g., `Rectangle('<variable_name>')`).
        """
        repr_str: str = f"Rectangle('{self.variable_name}')"
        return repr_str

    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Rectangle):
            from apysc._display.stage import get_stage_variable_name
            variable_name: str = self.variable_name
            stage_variable_name: str = get_stage_variable_name()
            expression: str = (
                f'var {variable_name} = {stage_variable_name}'
                f'\n  .rect({self.width.variable_name}, '
                f'{self.height.variable_name})'
            )
            attrs_str: str = self._make_rect_attrs_expression()
            expression += f'{attrs_str};'
            ap.append_js_expression(expression=expression)

    def _make_rect_attrs_expression(self) -> str:
        """
        Make rectangle attributes expression string.

        Returns
        -------
        rect_attrs_expression : str
            Rectangle attributes expression string.
        """
        from apysc._display import graphics_expression
        from apysc._display.graphics import Graphics
        graphics: Graphics = self.parent_graphics
        INDENT_NUM: int = 2
        rect_attrs_expression: str = '\n  .attr({'
        rect_attrs_expression = graphics_expression.append_fill_expression(
            graphics=graphics, expression=rect_attrs_expression,
            indent_num=INDENT_NUM)
        rect_attrs_expression = \
            graphics_expression.append_fill_opacity_expression(
                graphics=graphics, expression=rect_attrs_expression,
                indent_num=INDENT_NUM)
        rect_attrs_expression = graphics_expression.append_stroke_expression(
            graphics=graphics, expression=rect_attrs_expression,
            indent_num=INDENT_NUM)
        rect_attrs_expression = \
            graphics_expression.append_stroke_width_expression(
                graphics=graphics, expression=rect_attrs_expression,
                indent_num=INDENT_NUM)
        rect_attrs_expression = \
            graphics_expression.append_stroke_opacity_expression(
                graphics=graphics, expression=rect_attrs_expression,
                indent_num=INDENT_NUM)
        rect_attrs_expression = graphics_expression.append_x_expression(
            graphic=self, expression=rect_attrs_expression,
            indent_num=INDENT_NUM)
        rect_attrs_expression = graphics_expression.append_y_expression(
            graphic=self, expression=rect_attrs_expression,
            indent_num=INDENT_NUM)

        rect_attrs_expression += '\n  })'
        return rect_attrs_expression



"""Class implementation of floating point number.
"""

from typing import Any
from typing import Union

from apysc._type.number_value_interface import NumberValueInterface


class Number(NumberValueInterface[float, 'Number']):
    """
    Floating point number class for the apysc library.

    Notes
    -----
    Float class is the alias of the Number, and it behaves
    the same as the Number class.

    References
    ----------
    - Int and Number document
        - https://simon-ritchie.github.io/apysc/int_and_number.html
    - Int and Number common arithmetic operations document
        - https://bit.ly/3evzcVj
    - Int and Number common comparison operations document
        - https://bit.ly/3zolw6T
    """

    def __init__(self, value: Union[int, float, Any]) -> None:
        """
        Floating point number class for apysc library.

        Parameters
        ----------
        value : int or float or Int or Number
            Initial floating point number value. If int or Int value
            is specified, that value will be cast to float.

        Notes
        -----
        Float class is the alias of the Number, and it behaves
        the same as the Number class.

        References
        ----------
        - Int and Number document
            - https://simon-ritchie.github.io/apysc/int_and_number.html
        - Int and Number common arithmetic operations document
            - https://bit.ly/3evzcVj
        - Int and Number common comparison operations document
            - https://bit.ly/3zolw6T

        Examples
        --------
        >>> import apysc as ap
        >>> number: ap.Number = ap.Number(10.5)
        >>> number
        Number(10.5)

        >>> number == 10.5
        Boolean(True)

        >>> number == ap.Number(10.5)
        Boolean(True)

        >>> number >= 10.5
        Boolean(True)

        >>> number += 10.3
        >>> number
        Number(20.8)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Number):
            from apysc._converter import cast
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            with TemporaryNotHandlerScope():
                TYPE_NAME: str = var_names.NUMBER
                self.variable_name = expression_variables_util.\
                    get_next_variable_name(type_name=TYPE_NAME)
                super(Number, self).__init__(value=value, type_name=TYPE_NAME)
                self._value = cast.to_float_from_int(int_or_float=self.value)
                self.append_constructor_expression()

    def _set_value_and_skip_expression_appending(
            self, *, value: Union[int, float, Any]) -> None:
        """
        Update value attribute and skip expression appending.

        Parameters
        ----------
        value : int or float or Int or Number
            Any number value to set. If float or Number value is specified,
            that value will be cast to integer.
        """
        from apysc._converter import cast
        from apysc._validation import number_validation
        number_validation.validate_num(num=value)
        if isinstance(value, NumberValueInterface):
            value._value = cast.to_float_from_int(int_or_float=value._value)
            value_ = value._value
        else:
            value = cast.to_float_from_int(int_or_float=value)
            value_ = value
        self._value = value_

    def __repr__(self) -> str:
        """
        Get a representation string of this instance.

        Returns
        -------
        repr_str : str
            Representation string of this instance.
        """
        if not hasattr(self, '_value'):
            repr_str: str = 'Number(0)'
        else:
            repr_str = f'Number({self._value})'
        return repr_str



"""Class implementation for array.
"""

from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

from apysc._event.custom_event_interface import CustomEventInterface
from apysc._type.boolean import Boolean
from apysc._type.copy_interface import CopyInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface
from apysc._type.string import String
from apysc._type.variable_name_interface import VariableNameInterface

T = TypeVar('T')


class Array(
        CopyInterface, RevertInterface, CustomEventInterface, Generic[T]):
    """
    Array class for the apysc library.

    References
    ----------
    - Array document
        - https://simon-ritchie.github.io/apysc/array.html
    - Array class comparison interfaces document
        - https://simon-ritchie.github.io/apysc/array_comparison.html
    """

    _initial_value: Union[List[Any], tuple, 'Array']
    _value: List[Any]

    def __init__(
            self,
            value: Union[List[T], tuple, range, 'Array']) -> None:
        """
        Array class for the apysc library.

        Parameters
        ----------
        value : list or tuple or range or Array
            Initial array value.

        References
        ----------
        - Array document
            - https://simon-ritchie.github.io/apysc/array.html
        - Array class comparison interfaces document
            - https://simon-ritchie.github.io/apysc/array_comparison.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr
        Array([1, 2, 3])

        >>> arr[0]
        1

        >>> arr[1]
        2

        >>> arr = ap.Array((4, 5, 6))
        >>> arr
        Array([4, 5, 6])

        >>> arr = ap.Array(range(3))
        >>> arr
        Array([0, 1, 2])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            with TemporaryNotHandlerScope():
                TYPE_NAME: str = var_names.ARRAY
                self._validate_acceptable_value_type(value=value)
                value = self._convert_range_to_list(value=value)
                value_: Union[List[Any], tuple, 'Array'] = value
                self._initial_value = value_
                self._type_name = TYPE_NAME
                self._value = self._get_list_value(value=value)
                self.variable_name = expression_variables_util.\
                    get_next_variable_name(type_name=TYPE_NAME)
                self._append_constructor_expression()

    def _convert_range_to_list(
            self, *,
            value: Any) -> Union[List[Any], tuple, 'Array']:
        """
        Convert argument value to list that if specified
        value is range type.

        Parameters
        ----------
        value : list or tuple or range or Array
            Target value.

        Returns
        -------
        value : list or tuple or Array
            Converted value.
        """
        if isinstance(value, range):
            return list(value)
        return value

    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            expression: str = f'var {self.variable_name} = '
            if isinstance(self._initial_value, Array):
                expression += f'{self._initial_value.variable_name};'
            else:
                value_str: str = value_util.get_value_str_for_expression(
                    value=self._value)
                expression += f'{value_str};'
            ap.append_js_expression(expression=expression)

    def _get_list_value(
            self, *, value: Union[List[Any], tuple, 'Array']) -> List[Any]:
        """
        Get a list value from specified list, tuple, or Array value.

        Parameters
        ----------
        value : list or tuple or Array
            Specified list, tuple, or Array value.

        Returns
        -------
        list_val : list
            Converted list value.
        """
        if isinstance(value, tuple):
            return list(value)
        if isinstance(value, Array):
            return value._value
        return value

    def _validate_acceptable_value_type(
            self, *,
            value: Union[List[Any], tuple, range, 'Array']) -> None:
        """
        Validate that specified value is acceptable type or not.

        Parameters
        ----------
        value : list or tuple or range or Array
            Iterable value to check.

        Raises
        ------
        ValueError
            If specified value's type is not list, tuple, or Array.
        """
        if isinstance(value, (list, tuple, range, Array)):
            return
        raise ValueError(
            'Not acceptable value type is specified.'
            f'\nSpecified value type: {type(value)}'
            '\nAcceptable types: list, tuple, range, and Array')

    @property
    def value(self) -> Union[List[Any], tuple, 'Array']:
        """
        Get a current array value.

        Returns
        -------
        value : list
            Current array value.

        References
        ----------
        - apysc basic data classes common value interface
            - https://bit.ly/3Be1aij

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.value = [4, 5, 6]
        >>> arr.value
        [4, 5, 6]
        """
        return self._value

    @value.setter
    def value(self, value: Union[List[Any], tuple, 'Array']) -> None:
        """
        Set array value.

        Parameters
        ----------
        value : list or tuple or Array
            Iterable value (list, tuple, or Array) to set.

        References
        ----------
        apysc basic data classes common value interface
            https://bit.ly/3Be1aij
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='value', locals_=locals(),
                module_name=__name__, class_=Array):
            self._validate_acceptable_value_type(value=value)
            self._value = self._get_list_value(value=value)
            self._append_value_setter_expression(value=value)

    def _append_value_setter_expression(
            self, *, value: Union[List[Any], tuple, 'Array']) -> None:
        """
        Append value's setter expression.

        Parameters
        ----------
        value : list or tuple or Array
            Iterable value (list, tuple, or Array) to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_value_setter_expression,
                locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            expression: str = f'{self.variable_name} = '
            if isinstance(value, Array):
                expression += f'{value.variable_name};'
            else:
                value_str: str = value_util.get_value_str_for_expression(
                    value=value)
                expression += f'{value_str};'
            ap.append_js_expression(expression=expression)

    def append(self, value: T) -> None:
        """
        Add any value to the end of this array.
        This behaves same as push method.

        Parameters
        ----------
        value : *
            Any value to append.

        References
        ----------
        - Array class append and push interfaces document
            - https://simon-ritchie.github.io/apysc/array_append_and_push.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.append(4)
        >>> arr
        Array([1, 2, 3, 4])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.append, locals_=locals(),
                module_name=__name__, class_=Array):
            self._value.append(value)
            self._append_push_and_append_expression(value=value)

    def push(self, value: T) -> None:
        """
        Add any value to the end of this array.
        This behaves same as append method.

        Parameters
        ----------
        value : *
            Any value to append.

        References
        ----------
        - Array class append and push interfaces document
            - https://simon-ritchie.github.io/apysc/array_append_and_push.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.push(4)
        >>> arr
        Array([1, 2, 3, 4])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.push, locals_=locals(),
                module_name=__name__, class_=Array):
            self.append(value=value)

    def _append_push_and_append_expression(
            self, *, value: T) -> None:
        """
        Append push and append method expression.

        Parameters
        ----------
        value : *
            Any value to append.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_push_and_append_expression,
                locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression: str = (
                f'{self.variable_name}.push({value_str});'
            )
            ap.append_js_expression(expression=expression)

    def extend(self, other_arr: Union[List[T], tuple, 'Array']) -> None:
        """
        Concatenate argument array to this one. Argument array's
        values will positioned after this array's values.
        This method is similar to concat method, but there is a
        difference in whether the same variable will be
        updated (extend) or returned as a different variable (concat).

        Parameters
        ----------
        other_arr : list or tuple or Array
            Other array-like value to concatenate.

        References
        ----------
        - Array class extend and concat interfaces document
            - https://bit.ly/3r1TdIu

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.extend([4, 5, 6])
        >>> arr
        Array([1, 2, 3, 4, 5, 6])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.extend, locals_=locals(),
                module_name=__name__, class_=Array):
            self._validate_acceptable_value_type(value=other_arr)
            if isinstance(other_arr, Array):
                self._value.extend(other_arr.value)  # type: ignore
            else:
                self._value.extend(other_arr)
            self._append_extend_expression(other_arr=other_arr)

    def _append_extend_expression(
            self, *, other_arr: Union[List[T], tuple, 'Array']) -> None:
        """
        Append extend method expression.

        Parameters
        ----------
        other_arr : list or tuple or Array
            Other array-like value to concatenate.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_extend_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=other_arr)
            expression: str = (
                f'{self.variable_name} = {self.variable_name}'
                f'.concat({value_str});')
            ap.append_js_expression(expression=expression)

    def concat(
            self, other_arr: Union[List[T], tuple, 'Array']) -> 'Array':
        """
        Concatenate arugment array to this one. Argument array's
        values will positioned after this array's values.
        This method is similar to extend method, but there is a
        difference in whether the same variable will be
        updated (extend) or returned as a different variable (concat).

        Parameters
        ----------
        other_arr : list or tuple or Array
            Other array-like value to concatenate.

        Returns
        -------
        concatenated : Array
            Concatenated array value.

        References
        ----------
        - Array class extend and concat interfaces document
            - https://bit.ly/3r1TdIu

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr = arr.concat([4, 5, 6])
        >>> arr
        Array([1, 2, 3, 4, 5, 6])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.concat, locals_=locals(),
                module_name=__name__, class_=Array):
            concatenated: Array = self._copy()
            concatenated.extend(other_arr)
            self._append_concat_expression(
                concatenated=concatenated, other_arr=other_arr)
            return concatenated

    def _append_concat_expression(
            self, *, concatenated: VariableNameInterface,
            other_arr: Union[List[T], tuple, 'Array']) -> None:
        """
        Append concat method expression.

        Parameters
        ----------
        concatenated : Array
            Concatenated array value.
        other_arr : list or tuple or Array
            Other array-like value to concatenate.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_concat_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=other_arr)
            expression: str = (
                f'var {concatenated.variable_name} = '
                f'{self.variable_name}.concat({value_str});'
            )
            ap.append_js_expression(expression=expression)

    def insert(
            self, index: Union[int, Int], value: T) -> None:
        """
        Insert value to this array at a specified index.
        This behaves same as insert_at method.

        Parameters
        ----------
        index : int or Int
            Index to append value to.
        value : *
            Any value to append.

        References
        ----------
        - Array class insert and insert_at interfaces document
            - https://bit.ly/3G9LBtQ

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 3])
        >>> arr.insert(index=1, value=2)
        >>> arr
        Array([1, 2, 3])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.insert, locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._validation import number_validation
            number_validation.validate_integer(integer=index)
            if isinstance(index, ap.Int):
                index_: int = int(index.value)
            else:
                index_ = index
            value_: Any
            if isinstance(value, ap.Int):
                value_ = int(value.value)
            else:
                value_ = value
            self._value.insert(index_, value_)
            self._append_insert_expression(index=index, value=value)

    def insert_at(self, index: Union[int, Int], value: T) -> None:
        """
        Insert value to this array at a specified index.
        This behaves same as insert method.

        Parameters
        ----------
        index : int or Int
            Index to append value to.
        value : *
            Any value to append.

        References
        ----------
        - Array class insert and insert_at interfaces document
            - https://bit.ly/3G9LBtQ

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 3])
        >>> arr.insert_at(index=1, value=2)
        >>> arr
        Array([1, 2, 3])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.insert_at, locals_=locals(),
                module_name=__name__, class_=Array):
            self.insert(index=index, value=value)

    def _append_insert_expression(
            self, *, index: Union[int, Int], value: T) -> None:
        """
        Append insert method expression.

        Parameters
        ----------
        index : int or Int
            Index to append value to.
        value : *
            Any value to append.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_insert_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            index_str: str = value_util.get_value_str_for_expression(
                value=index)
            expression: str = (
                f'{self.variable_name}.splice({index_str}, 0, {value_str});'
            )
            ap.append_js_expression(expression=expression)

    def pop(self) -> T:
        """
        Remove this array's last value and return it.

        Returns
        -------
        value : *
            Removed value.

        References
        ----------
        - Array class pop interface document
            - https://simon-ritchie.github.io/apysc/array_pop.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> popped_val: int = arr.pop()
        >>> popped_val
        3

        >>> arr
        Array([1, 2])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.pop, locals_=locals(),
                module_name=__name__, class_=Array):
            value: T = self._value.pop()
            self._append_pop_expression(value=value)
            return value

    def _append_pop_expression(self, *, value: T) -> None:
        """
        Append pop method expression.

        Parameters
        ----------
        value : *
            Removed value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_pop_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            expression: str = f'{self.variable_name}.pop();'
            if isinstance(value, VariableNameInterface):
                expression = f'{value.variable_name} = {expression}'
            ap.append_js_expression(expression=expression)

    def remove(self, value: T) -> None:
        """
        Remove specified value from this array.

        Parameters
        ----------
        value : Any
            Value to remove.

        References
        ----------
        - Array class remove and remove_at interfaces document
            - https://bit.ly/3zDO6Cl

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 3, 5])
        >>> arr.remove(3)
        >>> arr
        Array([1, 5])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.remove, locals_=locals(),
                module_name=__name__, class_=Array):
            self._value.remove(value)
            self._append_remove_expression(value=value)

    def _append_remove_expression(self, *, value: T) -> None:
        """
        Append remove method expression.

        Parameters
        ----------
        value : Any
            Value to remove.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_remove_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._type import value_util
            index_var_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.INDEX)
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression: str = (
                f'var {index_var_name} = _.indexOf'
                f'({self.variable_name}, {value_str});'
                f'\n{self.variable_name}.splice({index_var_name}, 1);')
            ap.append_js_expression(expression=expression)

    def remove_at(self, index: Union[int, Int]) -> None:
        """
        Remove specified index value from this array.

        Parameters
        ----------
        index : int or Int
            Index to remove value.

        References
        ----------
        - Array class remove and remove_at interfaces document
            - https://bit.ly/3zDO6Cl

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.remove_at(1)
        >>> arr
        Array([1, 3])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.remove_at, locals_=locals(),
                module_name=__name__, class_=Array):
            self._validate_index_type_is_int(index=index)
            if isinstance(index, ap.Int):
                index_: int = int(index.value)
            else:
                index_ = index
            if index_ in self._value:
                del self._value[index_]
            self._append_remove_at_expression(index=index)

    def _append_remove_at_expression(
            self, *, index: Union[int, Int]) -> None:
        """
        Append remove_at method expression.

        Parameters
        ----------
        index : int or Int
            Index to remove value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_remove_at_expression,
                locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            index_str: str = value_util.get_value_str_for_expression(
                value=index)
            expression: str = (
                f'{self.variable_name}.splice({index_str}, 1);'
            )
            ap.append_js_expression(expression=expression)

    def reverse(self) -> None:
        """
        Reverse this array in place.

        References
        ----------
        - Array class reverse interface document
            - https://simon-ritchie.github.io/apysc/array_reverse.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.reverse()
        >>> arr
        Array([3, 2, 1])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.reverse, locals_=locals(),
                module_name=__name__, class_=Array):
            self._value.reverse()
            self._append_reverse_expression()

    def _append_reverse_expression(self) -> None:
        """
        Append reverse method expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_reverse_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            expression: str = (
                f'{self.variable_name}.reverse();'
            )
            ap.append_js_expression(expression=expression)

    def sort(self, *, ascending: bool = True) -> None:
        """
        Sort this array in place.

        Parameters
        ----------
        ascending : bool, default True
            Sort by ascending or not. If False is specified,
            values will be descending.

        References
        ----------
        - Array class sort interface document
            - https://simon-ritchie.github.io/apysc/array_sort.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([3, 5, 1, 4, 2])
        >>> arr.sort()
        >>> arr
        Array([1, 2, 3, 4, 5])

        >>> arr.sort(ascending=False)
        >>> arr
        Array([5, 4, 3, 2, 1])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.sort, locals_=locals(),
                module_name=__name__, class_=Array):
            self._value.sort()
            self._append_sort_expression()
            if not ascending:
                self.reverse()

    def _append_sort_expression(self) -> None:
        """
        Append sort method expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_sort_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            expression: str = (
                f'{self.variable_name}.sort();'
            )
            ap.append_js_expression(expression=expression)

    def slice(
            self,
            *,
            start: Optional[Union[int, Int]] = None,
            end: Optional[Union[int, Int]] = None) -> 'Array':
        """
        Slice this array by specified start and end indexes.

        Parameters
        ----------
        start : int or Int or None, default None
            Slicing start index.
        end : int or Int or None, default None
            Slicing end index (this index will not be including).

        Returns
        -------
        sliced_arr : Array
            Sliced array.

        References
        ----------
        - Array class slice interface document
            - https://simon-ritchie.github.io/apysc/array_slice.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3, 4])
        >>> arr.slice(start=1, end=3)
        Array([2, 3])

        >>> arr.slice(start=1)
        Array([2, 3, 4])

        >>> arr.slice(end=2)
        Array([1, 2])
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.slice, locals_=locals(),
                module_name=__name__, class_=Array):
            if isinstance(start, ap.Int):
                start_: Optional[int] = int(start.value)
            else:
                start_ = start
            if isinstance(end, ap.Int):
                end_: Optional[int] = int(end.value)
            else:
                end_ = end
            sliced_arr: Array = self._copy()
            sliced_arr._value = self._value[slice(start_, end_)]
            self._append_slice_expression(
                sliced_arr=sliced_arr, start=start, end=end)
            return sliced_arr

    def _append_slice_expression(
            self,
            *,
            sliced_arr: VariableNameInterface,
            start: Optional[Union[int, Int]],
            end: Optional[Union[int, Int]]) -> None:
        """
        Append slice method expression.

        Parameters
        ----------
        sliced_arr : Array
            Sliced array.
        start : int or Int or None
            Slicing start index.
        end : int or Int or None
            Slicing end index.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_slice_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            if start is None:
                start = 0
            expression: str = (
                f'var {sliced_arr.variable_name} = '
                f'{self.variable_name}.slice('
                f'{start}'
            )
            if end is not None:
                expression += f', {end}'
            expression += ');'
            ap.append_js_expression(expression=expression)

    def __getitem__(self, index: Union[int, Int]) -> T:
        """
        Get a specified index single value.

        Parameters
        ----------
        index : int or Int
            Array's index to get value. Currently not supported tuple
            value (e.g., slicing).

        Returns
        -------
        value : *
            Specified index's value.

        Raises
        ------
        ValueError
            If specified index type is not int and Int.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__getitem__', locals_=locals(),
                module_name=__name__, class_=Array):
            self._validate_index_type_is_int(index=index)
            index_: int = self._get_builtin_int_from_index(index=index)
            value: Any
            if len(self._value) <= index:
                value = ap.AnyValue(None)
            else:
                value = self._value[index_]
            self._append_getitem_expression(index=index, value=value)
            return value

    def _get_builtin_int_from_index(
            self, *, index: Union[int, Int]) -> int:
        """
        Get Python builtin integer from index value.

        Parameters
        ----------
        index : int or Int
            Specified array's index.

        Returns
        -------
        builtin_int_index : int
            Python builtin integer index value.
        """
        import apysc as ap
        if isinstance(index, ap.Int):
            return int(index.value)
        return index

    def _validate_index_type_is_int(
            self, *, index: Union[int, Int]) -> None:
        """
        Validate whether index value type is int (or Int) or not.

        Parameters
        ----------
        index : int or Int
            Index value to check.

        Raises
        ------
        ValueError
            If index type is not int or Int type.
        """
        if isinstance(index, (int, Int)):
            return
        raise ValueError(
            'Currently indexing is only supported int or Int types.'
            ' If you need to slice array please use slice method.')

    def _append_getitem_expression(
            self, *, index: Union[int, Int],
            value: T) -> None:
        """
        Append __getitem__ expression.

        Parameters
        ----------
        index : int or Int
            Array's index to get value.
        value : *
            Specified index's value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_getitem_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            value_: VariableNameInterface
            if not isinstance(value, VariableNameInterface):
                value_ = ap.AnyValue(None)
            else:
                value_ = value
            index_str: str = value_util.get_value_str_for_expression(
                value=index)
            expression: str = (
                f'var {value_.variable_name} = '
                f'{self.variable_name}[{index_str}];'
            )
            ap.append_js_expression(expression=expression)

    def __setitem__(self, index: Union[int, Int], value: T) -> None:
        """
        Set value to a specified index.

        Parameters
        ----------
        index : int or Int
            Array's index to set value. Currently not supported tuple
            value (e.g., slicing).
        value : *
            Any value to set.

        Raises
        ------
        ValueError
            If specified index type is not int and Int.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__setitem__', locals_=locals(),
                module_name=__name__, class_=Array):
            self._validate_index_type_is_int(index=index)
            index_: int = self._get_builtin_int_from_index(index=index)
            self._value[index_] = value
            self._append_setitem_expression(index=index, value=value)

    def _append_setitem_expression(
            self, *, index: Union[int, Int], value: T) -> None:
        """
        Append __setitem__ method expression.

        Parameters
        ----------
        index : int or Int
            Array's index to set value. Currently not supported tuple
            value (e.g., slicing).
        value : *
            Any value to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_setitem_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            index_str: str = value_util.get_value_str_for_expression(
                value=index)
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression: str = (
                f'{self.variable_name}[{index_str}] = '
                f'{value_str};'
            )
            ap.append_js_expression(expression=expression)

    def __delitem__(self, index: Union[int, Int]) -> None:
        """
        Delete specified index value from this array.

        Parameters
        ----------
        index : int or Int
            Array's index to delete. Currently not supported tuple
            value (e.g., slicing).

        Raises
        ------
        ValueError
            If specified index type is not int and Int.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__delitem__', locals_=locals(),
                module_name=__name__, class_=Array):
            self.remove_at(index=index)

    @property
    def length(self) -> Int:
        """
        Get length of this array.

        Returns
        -------
        length : Int
            This array's length.

        References
        ----------
        - Array class length interface document
            - https://simon-ritchie.github.io/apysc/array_length.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.length
        Int(3)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='length', locals_=locals(),
                module_name=__name__, class_=Array):
            length: ap.Int = ap.Int(len(self._value))
            self._append_length_expression(length=length)
            return length

    def _append_length_expression(self, *, length: Int) -> None:
        """
        Append length method expression.

        Parameters
        ----------
        length : Int
            Created length Int variable.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_length_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            expression: str = (
                f'{length.variable_name} = {self.variable_name}.length;'
            )
            ap.append_js_expression(expression=expression)

    def __len__(self) -> None:
        """
        This method is disabled and can't use from Array instance.
        """
        raise Exception(
            'Array instance can\'t apply len function.'
            ' Please use length property instead.')

    def join(self, sep: Union[str, String]) -> String:
        """
        Join this array values with specified separator string.

        Parameters
        ----------
        sep : str or String
            Separator string.

        Returns
        -------
        joined : String
            Joined string.

        References
        ----------
        - Array class join interface document
            - https://simon-ritchie.github.io/apysc/array_join.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 2, 3])
        >>> arr.join(sep=', ')
        String('1, 2, 3')
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.join, locals_=locals(),
                module_name=__name__, class_=Array):
            if isinstance(sep, ap.String):
                sep_: str = sep._value
            else:
                sep_ = sep
            values_: List[Any] = [str(value) for value in self._value]
            joined: ap.String = ap.String(sep_.join(values_))
            self._append_join_expression(joined=joined, sep=sep)
            return joined

    def _append_join_expression(
            self, *, joined: String, sep: Union[str, String]) -> None:
        """
        Append join method expression.

        Parameters
        ----------
        joined : String
            Joined string.
        sep : str or String
            Separator string.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_join_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            sep_str: str = value_util.get_value_str_for_expression(value=sep)
            expression: str = (
                f'{joined.variable_name} = {self.variable_name}'
                f'.join({sep_str});'
            )
            ap.append_js_expression(expression=expression)

    def __str__(self) -> str:
        """
        String conversion method.

        Returns
        -------
        string : str
            Converted value string.
        """
        if not hasattr(self, '_value'):
            return '[]'
        return str(self._value)

    def __repr__(self) -> str:
        """
        Get a representation string of this instance.

        Returns
        -------
        repr_str : str
            Representation string of this instance.
        """
        if not hasattr(self, '_value'):
            repr_str: str = 'Array([])'
        else:
            repr_str = f'Array({self._value})'
        return repr_str

    def index_of(self, value: T) -> Int:
        """
        Search specified value's index and return it.

        Parameters
        ----------
        value : *
            Any value to search.

        Returns
        -------
        index : Int
            Found position of index. If value is not contains,
            -1 will be returned.

        References
        ----------
        - Array class index_of interface document
            - https://simon-ritchie.github.io/apysc/array_index_of.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array([1, 3, 5])
        >>> arr.index_of(3)
        Int(1)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self.index_of, locals_=locals(),
                module_name=__name__, class_=Array):
            index: ap.Int = ap.Int(-1)
            try:
                index_: int = self._value.index(value)
            except Exception:
                index_ = -1
            index._value = index_
            self._append_index_of_expression(index=index, value=value)
            return index

    def _append_index_of_expression(
            self, *, index: Int, value: T) -> None:
        """
        Append index_of method expression.

        Parameters
        ----------
        index : Int
            Found position of index. If value is not contains,
            -1 will be set.
        value : *
            Any value to search.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_index_of_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            from apysc._type import value_util
            value_str: str = value_util.get_value_str_for_expression(
                value=value)
            expression: str = (
                f'{index.variable_name} = {self.variable_name}'
                f'.indexOf({value_str});'
            )
            ap.append_js_expression(expression=expression)

    def __eq__(self, other: Any) -> Any:
        """
        Equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. list or Array types are acceptable.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__eq__', locals_=locals(),
                module_name=__name__, class_=Array):
            result: ap.Boolean
            if isinstance(other, Array):
                result = ap.Boolean(self._value == other._value)
            else:
                result = ap.Boolean(self._value == other)
                other = self._convert_other_val_to_array(other=other)
            if isinstance(other, VariableNameInterface):
                self._append_eq_expression(result=result, other=other)
            return result

    def _convert_other_val_to_array(self, *, other: Any) -> Any:
        """
        If comparison's other value is list value, then convert it to
        Array instance.

        Parameters
        ----------
        other : *
            Other value to compare.

        Returns
        -------
        converted_val : *
            Converted value. If other value is list, then this will
            be Array type. Otherwise this will be returned directly
            (not to be converted).
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._convert_other_val_to_array, locals_=locals(),
                module_name=__name__, class_=Array):
            if isinstance(other, list):
                return Array(other)
            return other

    def _append_eq_expression(
            self, *, result: Boolean,
            other: VariableNameInterface) -> None:
        """
        Append an __eq__ expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : Array
            Array other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_eq_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            expression: str = (
                f'{result.variable_name} = '
                f'_.isEqual({self.variable_name}, {other.variable_name});'
            )
            ap.append_js_expression(expression=expression)

    def __ne__(self, other: Any) -> Any:
        """
        Not equal comparison method.

        Parameters
        ----------
        other : *
            Other value to compare. list or Array types are acceptable.

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__ne__', locals_=locals(),
                module_name=__name__, class_=Array):
            other = self._convert_other_val_to_array(other=other)
            result: ap.Boolean = self == other
            result = result.not_
            if isinstance(other, VariableNameInterface):
                self._append_ne_expression(result=result, other=other)
            return result

    def _append_ne_expression(
            self, *, result: Boolean,
            other: VariableNameInterface) -> None:
        """
        Append a __ne__ expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : Array
            Array other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ne_expression, locals_=locals(),
                module_name=__name__, class_=Array):
            expression: str = (
                f'{result.variable_name} = '
                f'!_.isEqual({self.variable_name}, {other.variable_name});'
            )
            ap.append_js_expression(expression=expression)

    def __bool__(self) -> bool:
        """
        Get a boolean value whether this array is empty or not.

        Returns
        -------
        result : bool
            If this array is empty, True will be returned.
        """
        return bool(self._value)

    _value_snapshots: Dict[str, List[T]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make values' snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_value_snapshots',
            value=[*self._value], snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert values if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._value = self._value_snapshots[snapshot_name]



"""The module that implements PyPI settings.
"""

from setuptools import find_packages
from setuptools import setup

from apysc import __version__

_DESCRIPTION: str = (
    'apysc is a Python\'s frontend library to create '
    'html and js file, that has the ActionScript 3 (as3)-like interface.'
)

_LONG_DESCRIPTION: str = (
    'apysc is the Python\'s frontend library to '
    'create html and js file, that has a ActionScript 3 '
    '(as3)-like interface.'
    ' For more details, please see GitHub repository:'
    ' https://github.com/simon-ritchie/apysc'
)

setup(
    name='apysc',
    version=__version__,
    url='https://github.com/simon-ritchie/apysc',
    maintainer='simon-ritchie',
    maintainer_email='',
    description=_DESCRIPTION,
    long_description=_LONG_DESCRIPTION,
    packages=find_packages(
        exclude=('tests', 'tests.*', 'test_projects', 'test_projects.*')),
    install_requires=[
        'typing-extensions',
        'html-minifier',
    ],
    include_package_data=True,
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: MIT License',
    ],
)



"""Class implementation for boolean.
"""

from typing import Any
from typing import Dict
from typing import Union

from apysc._event.custom_event_interface import CustomEventInterface
from apysc._type.copy_interface import CopyInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface
from apysc._type.variable_name_interface import VariableNameInterface


class Boolean(CopyInterface, RevertInterface, CustomEventInterface):
    """
    Boolean class for the apysc library.

    Notes
    -----
    The Bool class is the alias of the Boolean, and it behaves
    the same as the Boolean class.

    References
    ----------
    - Boolean document
        - https://simon-ritchie.github.io/apysc/boolean.html
    """

    _initial_value: Union[bool, int, Int, 'Boolean']
    _value: bool

    def __init__(self, value: Union[bool, int, Int, 'Boolean']) -> None:
        """
        Boolean class for apysc library.

        Notes
        -----
        The Bool class is the alias of the Boolean, and it behaves
        the same as the Boolean class.

        Parameters
        ----------
        value : bool or int or Boolean or Int
            Initial boolean value. 0 or 1 are acceptable for integer
            value.

        References
        ----------
        - Boolean document
            - https://simon-ritchie.github.io/apysc/boolean.html

        Examples
        --------
        >>> import apysc as ap
        >>> bool_val_1: ap.Boolean = ap.Boolean(True)
        >>> bool_val_1
        Boolean(True)

        >>> bool_val_2: ap.Bool = ap.Bool(True)
        >>> bool_val_2
        Boolean(True)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__init__', locals_=locals(),
                module_name=__name__, class_=Boolean):
            from apysc._expression import expression_variables_util
            from apysc._expression import var_names
            from apysc._expression.event_handler_scope import \
                TemporaryNotHandlerScope
            from apysc._validation import number_validation
            with TemporaryNotHandlerScope():
                TYPE_NAME: str = var_names.BOOLEAN
                number_validation.validate_int_is_zero_or_one(
                    integer=value)  # type: ignore
                self._initial_value = value
                value_: bool = self._get_bool_from_arg_value(value=value)
                self._value = value_
                self._type_name = TYPE_NAME
                self.variable_name = expression_variables_util.\
                    get_next_variable_name(type_name=TYPE_NAME)
                self._append_constructor_expression()

    def _get_bool_from_arg_value(
            self, *, value: Union[bool, int, Int, 'Boolean']) -> bool:
        """
        Get bool value from specified argument value.

        Parameters
        ----------
        value : bool or int or Boolean or Int
            Specified value. 0 or 1 are acceptable for integer
            value.

        Returns
        -------
        result : bool
            Converted boolean value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._get_bool_from_arg_value,
                locals_=locals(),
                module_name=__name__, class_=Boolean):
            from apysc._converter import cast
            from apysc._type.number_value_interface import NumberValueInterface
            from apysc._validation import bool_validation
            if isinstance(value, (int, float, NumberValueInterface)):
                result: bool = cast.to_bool_from_int(
                    integer=value)  # type: ignore
            elif isinstance(value, Boolean):
                result = value._value
            else:
                result = value
            bool_validation.validate_bool(value=result)
            return result

    def _append_constructor_expression(self) -> None:
        """
        Append constructor expression.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_constructor_expression,
                locals_=locals(),
                module_name=__name__, class_=Boolean):
            from apysc._type.variable_name_interface import \
                VariableNameInterface
            expression: str = f'var {self.variable_name} = '
            if isinstance(self._initial_value, VariableNameInterface):
                expression += f'Boolean({self._initial_value.variable_name});'
            elif self._value:
                expression += 'true;'
            else:
                expression += 'false;'
            ap.append_js_expression(expression=expression)

    @property
    def value(self) -> Union[bool, int, Int, 'Boolean']:
        """
        Get a current boolean value.

        Returns
        -------
        value : bool
            Current boolean value.

        References
        ----------
        - apysc basic data classes common value interface
            - https://bit.ly/3Be1aij

        Examples
        --------
        >>> import apysc as ap
        >>> bool_val: ap.Boolean = ap.Boolean(True)
        >>> bool_val.value = False
        >>> bool_val.value
        False

        >>> bool_val.value = ap.Boolean(True)
        >>> bool_val.value
        True
        """
        return self._value

    @value.setter
    def value(self, value: Union[bool, int, Int, 'Boolean']) -> None:
        """
        Set boolean value.

        Parameters
        ----------
        value : bool or int or Boolean or Int
            Any boolean value to set.

        References
        ----------
        apysc basic data classes common value interface
            https://bit.ly/3Be1aij
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='value', locals_=locals(),
                module_name=__name__, class_=Boolean):
            from apysc._type.variable_name_interface import \
                VariableNameInterface
            self._set_value_and_skip_expression_appending(value=value)
            if isinstance(value, VariableNameInterface):
                self._append_value_setter_expression(value=value)
            else:
                self._append_value_setter_expression(value=self._value)

    def _append_value_setter_expression(
            self, *, value: Union[bool, int, Int, 'Boolean']) -> None:
        """
        Append value's setter expression.

        Parameters
        ----------
        value : bool or VariableNameInterface
            Any value to set.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_value_setter_expression,
                locals_=locals(),
                module_name=__name__, class_=Boolean):
            from apysc._type.variable_name_interface import \
                VariableNameInterface
            expression: str = f'{self.variable_name} = '
            if isinstance(value, VariableNameInterface):
                expression += f'Boolean({value._variable_name});'
            elif value:
                expression += 'true;'
            else:
                expression += 'false;'
            ap.append_js_expression(expression=expression)

    def _set_value_and_skip_expression_appending(
            self, *, value: Union[bool, int, Int, 'Boolean']) -> None:
        """
        Update value attribute and skip expression appending.

        Parameters
        ----------
        value : bool or int or Boolean or Int
            Any boolean value to set.
        """
        value_: bool = self._get_bool_from_arg_value(value=value)
        self._value = value_

    def __bool__(self) -> bool:
        """
        Get a boolean value directly.

        Returns
        -------
        result : bool
            Current boolean value.
        """
        return self._value

    def __repr__(self) -> str:
        """
        Get a representation string of this instance.

        Returns
        -------
        repr_str : str
            Representation string of this instance.
        """
        if not hasattr(self, '_value'):
            repr_str: str = 'Boolean(False)'
        else:
            repr_str = f'Boolean({self._value})'
        return repr_str

    _value_snapshots: Dict[str, bool]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._set_single_snapshot_val_to_dict(
            dict_name='_value_snapshots',
            value=self._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._value = self._value_snapshots[snapshot_name]

    def __eq__(self, other: Any) -> Any:
        """
        Comparison method for equal condition.

        Parameters
        ----------
        other : *
            Other value to compare (Boolean, bool, int, or Int).

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__eq__', locals_=locals(),
                module_name=__name__, class_=Boolean):
            self._validate_comparison_other_type(other=other)
            result: Boolean
            if isinstance(other, Boolean):
                result = Boolean(self._value == other._value)
                self._append_eq_expression(result=result, other=other)
                return result
            elif isinstance(other, ap.Int):
                other_ = bool(other.value)
                result = Boolean(self._value == other_)
            else:
                other = bool(other)
                result = Boolean(self._value == other)
                other = Boolean(other)
            self._append_eq_expression(result=result, other=other)
            return result

    def _validate_comparison_other_type(self, *, other: Any) -> None:
        """
        Validate comparison's other value type.

        Parameters
        ----------
        other : *
            Other value to compare.

        Raises
        ------
        ValueError
            If other value type is not Boolean, bool, Int, and int.
        """
        import apysc as ap
        ACCEPTABLE_TYPES: tuple = (Boolean, bool, ap.Int, int)
        if isinstance(other, ACCEPTABLE_TYPES):
            return
        raise ValueError(
            'Can\'t acceptable comparison value type is specified: '
            f'{type(other)}, {other}'
            f'\nAcceptable value types are: {ACCEPTABLE_TYPES}')

    def _append_eq_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __eq__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : Boolean or Int
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_eq_expression, locals_=locals(),
                module_name=__name__, class_=Boolean):
            other_str: str = other.variable_name
            if isinstance(other, ap.Int):
                other_str = f'Boolean({other_str});'
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} === {other_str};'
            )
            ap.append_js_expression(expression=expression)

    def __ne__(self, other: Any) -> Any:
        """
        Comparison method for not equal condition.

        Parameters
        ----------
        other : *
            Other value to compare (Boolean, bool, int, or Int).

        Returns
        -------
        result : Boolean
            Comparison result.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='__ne__', locals_=locals(),
                module_name=__name__, class_=Boolean):
            result: Boolean = self == other
            result = result.not_
            if isinstance(other, (Boolean, ap.Int)):
                self._append_ne_expression(result=result, other=other)
            return result

    def _append_ne_expression(
            self, *, result: VariableNameInterface,
            other: VariableNameInterface) -> None:
        """
        Append __ne__ method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        other : Boolean or Int
            Other value to compare.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_ne_expression, locals_=locals(),
                module_name=__name__, class_=Boolean):
            other_str: str = other.variable_name
            if isinstance(other, ap.Int):
                other_str = f'Boolean({other_str});'
            expression: str = (
                f'{result.variable_name} = '
                f'{self.variable_name} !== {other_str};'
            )
            ap.append_js_expression(expression=expression)

    @property
    def not_(self) -> 'Boolean':
        """
        Get a not condition Boolean value.

        Returns
        -------
        result : Boolean
            Not condition Boolean value.

        Examples
        --------
        >>> import apysc as ap
        >>> bool_val: ap.Boolean = ap.Boolean(True)
        >>> bool_val.not_
        Boolean(False)

        >>> bool_val.value = False
        >>> bool_val.not_
        Boolean(True)
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_='not_', locals_=locals(),
                module_name=__name__, class_=Boolean):
            result: Boolean = Boolean(not self)
            self._append_not_prop_expression(result=result)
            return result

    def _append_not_prop_expression(
            self, *, result: VariableNameInterface) -> None:
        """
        Append not_ property expression.

        Parameters
        ----------
        result : Boolean
            Result Boolean value.
        """
        import apysc as ap
        with ap.DebugInfo(
                callable_=self._append_not_prop_expression,
                locals_=locals(),
                module_name=__name__, class_=Boolean):
            expression: str = (
                f'{result.variable_name} = '
                f'!{self.variable_name};'
            )
            ap.append_js_expression(expression=expression)
